import numpy as np
from ._unfold_numba import _bloch_unfold

def bloch_unfold(A, kpts, N, out=None):
    """Unfold matrices evaluated at k-points onto a supercell.

    The supercell (SC) consists of N repetitions of a smaller 
    building block described by the k-matrices. At the moment,
    unfolding is implemented only along two directions. This means
    that even though N is compsed of (3,) integers, only 2 of them
    can be greater than 1. Finally, note that the last dimension 
    != 1 is used in the innermost unfolding.
    
    """
    
    assert A.dtype == complex, 'Input arrays must be complex128!'

    i, j, k = N
    nr, nc = A.shape[1:]
    
    # NOTE that for 1D expansion only 2nd column of kpts are used
    if i == 1:
        if k == 1: # (1D) y-expand
            m, n = 1, j
            kpts = kpts[:, [0,1]]
        else: # (1D) z- or (2D) yz-expand
            m, n = j, k
            kpts = kpts[:, [1,2]]
    else: # i>1
        if j == 1:
            if k == 1: # (1D) x-expand
                m, n = 1, i
                kpts = kpts[:, [0,0]] # Hack to swap x-coord to last position.
            else: # (2D) xz-expand
                m, n = i, k
                kpts = kpts[:, [0,2]]
        else:
            if k==1: # xy-expand
                m, n = i, j
                kpts = kpts[:, [0,1]]
            else:
                raise NotImplementedError('Cannot expand 3D supercells.')

    mxn = m*n    
    if out is None:
        out = np.empty((mxn, nr, mxn, nc), complex)
    else:
        assert out.dtype ==complex, 'Output matrix must be complex128!'
        if out.ndim!=4:
            out.shape = (mxn, nr, mxn, nc)

    _bloch_unfold(A, kpts, m, n, out)
    
    out.shape = (mxn*nr,mxn*nc)
    return out



import numpy as np


def remove_pbc(basis, A_kMM, direction='x', eps=-1e-3):
    """Mask matrix elements above diagonal connecting neighbors cells
    in parallel direction.

    """

    # Transport direction
    p_dir = 'xyz'.index(direction)

    L = basis.atoms.cell[p_dir, p_dir]

    cutoff = L/2 - eps
    # Coordinates of central unit cell i (along transport)
    centers_p_i = basis.centers[:, p_dir]
    # Coordinates of neighbooring unit cell j
    # centers_p_j = centers_p_i + L
    # Distance between j atoms and i atoms
    dist_p_ji = np.abs(centers_p_i[:, None] - centers_p_i[None, :])
    # Mask j atoms farther than L
    mask_ji = (dist_p_ji < cutoff).astype(A_kMM.dtype)

    A_kMM *= mask_ji[None, :]


import numpy as np
from scipy import linalg as la
from .coupledhamiltonian import CoupledHamiltonian

class InternalSelfEnergy(CoupledHamiltonian):

    def __init__(self, hs_dii, hs_dim, selfenergies=[], eta=1e-5):
        # LeadSelfEnergy.__init__(self, hs_dii, (None, None), hs_dim, eta)
        self.h_ii, self.s_ii = hs_dii # onsite principal layer
        self.h_im, self.s_im = hs_dim # coupling to the central region
#        self.nbf_i = self.h_im.shape[0] # nbf_m for the internal self-energy
#        self.nbf_m = self.h_im.shape[1] # nbf_m for the scattering region
        self.eta = eta
        self.energy = None
        self.bias = 0
        #
        self.sigma_mm = np.empty((self.nbf_m, self.nbf_m), dtype=complex)
        self.Ginv = np.empty((self.nbf_i, self.nbf_i), dtype=complex)
        self.selfenergies = selfenergies

        CoupledHamiltonian.__init__(self, self.h_ii, self.s_ii, self.selfenergies)

    @property
    def nbf_i(self):
        return self.h_im.shape[0]

    @property
    def nbf_m(self):
        return self.h_im.shape[1]

    def align_bf(self, align_bf):
        diff = super().align_bf(align_bf)
        h_im = self.h_im
        s_im = self.s_im
        h_im -= diff * s_im

    def retarded(self, energy):
        """Return self-energy (sigma) evaluated at specified energy."""
        if energy != self.energy:
            self.energy = energy
            z = energy - self.bias + self.eta * 1.j
            tau_im = z * self.s_im - self.h_im
            a_im = np.linalg.solve(self.get_Ginv(energy), tau_im)
            tau_mi = z * self.s_im.T.conj() - self.h_im.T.conj()
            self.sigma_mm[:] = np.dot(tau_mi, a_im)

        return self.sigma_mm

    def get_lambda(self, energy):
        """Return the lambda (aka Gamma) defined by i(S-S^d).

        Here S is the retarded selfenergy, and d denotes the hermitian
        conjugate.
        """
        sigma_mm = self.retarded(energy)
        return 1.j * (sigma_mm - sigma_mm.T.conj())

    def get_Ginv(self, energy, inverse=True):

        z = energy - self.bias + self.eta * 1.j

        self.Ginv[:] = z
        self.Ginv *= self.S
        self.Ginv -= self.H

        for selfenergy in self.selfenergies:
            self.Ginv -= selfenergy.retarded(energy)

        if inverse:
            return self.Ginv
        else:
            return la.inv(self.Ginv)

    def apply_retarded(self, energy, X):
        """Apply retarded Green function to X.

        Returns the matrix product G^r(e) . X
        """
        return la.solve(self.get_Ginv(energy), X)

    def apply_overlap(self, energy, trace=False, diag=False):
        """Apply retarded Green function to S."""
        GS = self.apply_retarded(energy, self.S)
        if trace:
            return np.trace(GS)
        if diag:
            return GS.diagonal()
        return GS

    def dos(self, energy):
        """Total density of states -1/pi Im(Tr(GS))"""
        if self.S is None:
            return -self.get_Ginv(energy, inverse=False).imag.trace() / np.pi
        else:
            GS = self.apply_retarded(energy, self.S)
            return -GS.imag.trace() / np.pi

    def get_matsubara(self, beta, n):
        w_n = np.pi/beta * (2*n + 1)
        energy = 1.j * w_n - self.eta * 1.j
        return self.retarded(energy)

    ####### CONVENIENT ALISES ########

    @property
    def H(self):
        return self.h_ii
    @property
    def S(self):
        return self.s_ii
    @H.setter
    def H(self, H):
        self.h_ii = H
    @S.setter
    def S(self, H):
        self.s_ii = H

    ######## MODIFIERS ################

    def apply_rotation(self, c_mm):
        CoupledHamiltonian.apply_rotation(self, c_mm)
        self.h_im[:] = np.dot(c_mm.T.conj(), self.h_im)
        self.s_im[:] = np.dot(c_mm.T.conj(), self.s_im)

    def cutcoupling_bfs(self, bfs, apply=False):
        h_pp, s_pp = CoupledHamiltonian.cutcoupling_bfs(self, bfs, apply)
        if apply:
            for m in bfs:
                self.h_im[m, :] = 0.0
                self.s_im[m, :] = 0.0
        return h_pp, s_pp

    def take_bfs(self, bfs, apply=False):
        h_pp, s_pp, c_mm = CoupledHamiltonian.take_bfs(self, bfs, apply)
        if apply:
            self.h_im = np.dot(c_mm.T.conj(), self.h_im)
            self.s_im = np.dot(c_mm.T.conj(), self.s_im)
            self.Ginv = np.empty(self.H.shape, complex)
        return h_pp, s_pp, c_mm



import numpy as np

from ._recursive import get_G
from .unfold import bloch_unfold


class PrincipalLayer:
    """Principal Layer (PL) Green's function.
    
    Allows to compute a surface Green's Function for a supercell (SC)
    composed of a repeating principal layer (PL). The inputs are
    converged Hamiltonians and Overlaps for the PL evaluated at 
    tranverse k-points.
    
    """

    def __init__(self, kpts, hs_kii, hs_kij, Nr=None, eta=1e-5):
        """
        Args:
            kpts : np.ndarray (shape = (# of k-points, 3))
                tranverse k-points.
            hs_kii : (tuple, list)
                pointers to onsite Hamiltonians and Overlaps (at kpts).
            hs_kij : (tuple, list)
                pointers to coupling Hamiltonians and Overlaps (at kpts).
            Nr : tuple (size = (3,))
                # of transverse PLs.
                    
        """

        self.kpts = kpts
        self.h_kii, self.s_kii = hs_kii
        self.h_kij, self.s_kij = hs_kij
        self.Nr = Nr or self.Nk
        assert self.kpts.shape[1] == 3, "Invalid k-points. Must have (x,y,z) components."
        assert len(self.Nr) == 3, "Invalid # of transverse PLs. Must be (Nx, Ny, Nz)."

        self.h_ii = self.bloch_unfold(self.h_kii) # onsite supercell.
        self.s_ii = self.bloch_unfold(self.s_kii) # coupling to neighbor supercell.

        self.energy = None
        self.bias = 0.
        self.eta = eta

        self.G_kii = np.empty((kpts.shape[0],) + 2*(self.h_kii.shape[1],), complex)
        self.G = np.empty(2*(self.nbf_i,), complex) 

    @property
    def nbf_i(self):
        """# of supercell basis functions."""
        return self.h_ii.shape[0]

    @property
    def H(self): # Alias
        return self.h_ii

    @property
    def S(self): # Alias
        return self.s_ii

    @property
    def Nk(self):
        """# of transverse k-points"""
        return np.array([len(set(self.kpts[:,i])) for i in range(3)])

    def get_G(self, energy):
        """Get retarded Green function at specified energy."""
        if energy != self.energy:
            self.energy = energy
            get_G(self.G_kii, self.h_kii, self.s_kii, self.h_kij, 
                    self.s_kij, self.energy, self.eta, self.bias)
            self.bloch_unfold(self.G_kii, self.G)
        return self.G

    def bloch_unfold(self, A, out=None):
        """Unfold to supercell matrix.
        
        Args:
            A : Matrices at transverse k-points.
            out : (optional) supercell matrix 
        """
        return bloch_unfold(A, self.kpts, self.Nr, out)

    def dos(self, energy):
        """Total density of states -1/pi Im(Tr(GS))"""
        G = self.get_G(energy)
        return - np.dot(G, self.S).imag.trace() / np.pi

    def pdos(self, energy):
        """Projected density of states -1/pi Im(SGS/S)"""
        G = self.get_G(energy)
        SGS = np.linag.multi_dot([self.S, G, self.S])
        return - (SGS.diagonal() / self.S.diagonal()).imag / np.pi


class PrincipalSelfEnergy(PrincipalLayer):
    """Principal Layer (PL) Self-energy.
    
    Allows to compute a surface self-energy for a supercell (SC)
    composed of a repeating principal layer (PL). The inputs are
    converged Hamiltonians and Overlaps for the PL evaluated at 
    tranverse k-points.
    

    TODO :: order return!
    """

    ids = ['left','right']

    def __init__(self, kpts, hs_kii, hs_kij, Nr=None, eta=1e-5, id='left'):

        assert id in PrincipalSelfEnergy.ids, f"Invalid id for Principal self-energy. Choose between {PrincipalSelfEnergy.ids}"
        self.id = id
        if id == 'right':
            h_kij = hs_kij[0].swapaxes(1,2).conj()
            s_kij = hs_kij[1].swapaxes(1,2).conj()
            hs_kij = (h_kij, s_kij)

        super().__init__(kpts, hs_kii, hs_kij, Nr, eta)

        self.h_im = self.bloch_unfold(self.h_kij)
        self.s_im = self.bloch_unfold(self.s_kij)

        self.sigma_mm = np.empty(2*(self.nbf_i,), complex)

    def retarded(self, energy):
        """Return self-energy (sigma) evaluated at specified energy."""
        if energy != self.energy:
            z = energy - self.bias + self.eta * 1.j
            tau_im = z * self.s_im - self.h_im
            G = self.get_G(energy)
            tau_mi = z * self.s_im.T.conj() - self.h_im.T.conj()
            tau_mi.dot(G).dot(tau_im, out=self.sigma_mm)

        return self.sigma_mm

    def get_lambda(self, energy):
        """Return the lambda (aka Gamma) defined by i(S-S^d).

        Here S is the retarded selfenergy, and d denotes the hermitian
        conjugate.
        """
        sigma_mm = self.retarded(energy)
        return 1.j * (sigma_mm - sigma_mm.T.conj())


import numpy as np
import cupy as cp

device = cp.cuda.device

cublas = device.cublas
cusolver = device.cusolver

cublas_handle = device.get_cublas_handle()
cusolver_handle = device.get_cusolver_handle()

def _assert_farray(*arrays):
    for array in arrays:
        assert array.flags.f_contiguous


def solve(a, b, overwrite_b=False):
    _assert_farray(a, b)
    if overwrite_b:
        b = b.copy('A')
    
    getrf = cusolver.zgetrf
    getrf_buffer = cusolver.zgetrf_bufferSize
    getrs = cusolver.zgetrs
    trans = cublas.CUBLAS_OP_N

    m, n = a.shape
    nrhs = b.shape[1]
    lda = m
    ldb = m

    dev_info = cp.empty(1, dtype=np.int32)
    ipiv = cp.empty(m, np.int32)

    buffersize = getrf_buffer(cusolver_handle, m, n, a.data.ptr, lda)
    workspace = cp.empty(buffersize, dtype=a.dtype)

    getrf(cusolver_handle, m, n, a.data.ptr, lda, workspace.data.ptr, ipiv.data.ptr, dev_info.data.ptr)
    del workspace
    
    getrs(cusolver_handle, trans, n, nrhs, a.data.ptr, lda, ipiv.data.ptr, b.data.ptr, ldb, dev_info.data.ptr)
    
    return b


def lrsolve(a, b):
    _assert_farray(a, b)
    
    getrf = cusolver.zgetrf
    getrf_buffer = cusolver.zgetrf_bufferSize
    getrs = cusolver.zgetrs
    trans = cublas.CUBLAS_OP_N

    m, n = a.shape
    nrhs = b.shape[1]
    lda = m
    ldb = m

    dev_info = cp.empty(1, dtype=np.int32)
    ipiv = cp.empty(m, np.int32)

    buffersize = getrf_buffer(cusolver_handle, m, n, a.data.ptr, lda)
    workspace = cp.empty(buffersize, dtype=a.dtype)

    getrf(cusolver_handle, m, n, a.data.ptr, lda, workspace.data.ptr, ipiv.data.ptr, dev_info.data.ptr)
    del workspace
    
    getrs(cusolver_handle, trans, n, nrhs, a.data.ptr, lda, ipiv.data.ptr, b.data.ptr, ldb, dev_info.data.ptr)
    
    cp.conj(a, out=a)

    c = b.copy(order='C')

    getrs(cusolver_handle, trans, n, nrhs, a.data.ptr, lda, ipiv.data.ptr, c.data.ptr, ldb, dev_info.data.ptr)
    
    return c


def tsolve(a, b, c):
    _assert_farray(a, b, c)
    
    getrf = cusolver.zgetrf
    getrf_buffer = cusolver.zgetrf_bufferSize
    getrs = cusolver.zgetrs
    trans = cublas.CUBLAS_OP_N

    m, n = a.shape
    nrhs = b.shape[1]
    lda = m
    ldb = m

    dev_info = cp.empty(1, dtype=np.int32)
    ipiv = cp.empty(m, np.int32)

    buffersize = getrf_buffer(cusolver_handle, m, n, a.data.ptr, lda)
    workspace = cp.empty(buffersize, dtype=a.dtype)

    getrf(cusolver_handle, m, n, a.data.ptr, lda, workspace.data.ptr, ipiv.data.ptr, dev_info.data.ptr)
    del workspace
    
    getrs(cusolver_handle, trans, n, nrhs, a.data.ptr, lda, ipiv.data.ptr, b.data.ptr, ldb, dev_info.data.ptr)
    
    trans = cublas.CUBLAS_OP_C

    getrs(cusolver_handle, trans, n, nrhs, a.data.ptr, lda, ipiv.data.ptr, c.data.ptr, ldb, dev_info.data.ptr)
    
    return b, c


#!/usr/bin/env python

import numpy as np
from pathlib import Path

from scipy.sparse import csr_matrix, coo_matrix
from ase.units import Hartree

def read_bincsr_matrix(filename, dtype=np.float64):
    dt = np.dtype([
        ('o1', np.uint32),
        ('x', np.uint32),
        ('y', np.uint32),
        ('real', dtype),
        ('imag', dtype),
        ('o2', np.uint32)])
    
    coo = np.fromfile(filename, dt, offset=0)
    data = coo['real'] + 1.j * coo['imag']
    mat = coo_matrix((data,(coo['x']-1,coo['y']-1))).toarray()
    tri2full(mat,UL='U')
    return mat


cc_path = Path('/usr/scratch/mont-fort18/ggandus/k-point/sc')


get_ipython().system('egrep BRILLOUIN /usr/scratch/mont-fort18/ggandus/k-point/sc/energy.out')


# ## PREPROCESS SCATTERING
# 
#     i) read in reversed k-points (convention uses +$k_x$)
#     ii) transpose matrix A[k]=A[-k].T

h_cc_k_files = [cc_path/f'sc-KS_SPIN_1_K_{i}-1_0.csr' for i in range(8,0,-1)]
s_cc_k_files = [cc_path/f'sc-S_SPIN_1_K_{i}-1_0.csr' for i in range(8,0,-1)]


read_bincsr = lambda file: read_bincsr_matrix(file).T.copy('C')
h_cc_k = np.stack(list(map(read_bincsr, h_cc_k_files)))
s_cc_k = np.stack(list(map(read_bincsr, s_cc_k_files)))

plt.spy(h_cc_k[-1])

atoms_cc = read(cc_path/'sc.xyz')
basis_cc = basis.Basis.from_dictionary(atoms_cc, {'C':13,'Si':13,'O':13})

get_ipython().system('egrep -i fermi /usr/scratch/mont-fort18/ggandus/k-point/sc/energy.out')

fermi_cc = 0.26936888972782 * Hartree
fermi_cc

h_cc_k *= Hartree
h_cc_k -= fermi_cc * s_cc_k

kpts_t, h_cc_kii, s_cc_kii, h_cc_kij, s_cc_kij = tools.prepare_leads_matrices(basis_cc, h_cc_k, s_cc_k, (4,4,1))

del h_cc_k
del s_cc_k


# In[24]:


kpts_t


# ### SAVE

# In[ ]:


np.save('hs_cc_ii_ij',(h_cc_kii, s_cc_kii, h_cc_kij, s_cc_kij))


# In[22]:


plt.spy(h_cc_kij[0])


# In[23]:


pl_path = cc_path.parent/'uc'#Path('./uc/')


# In[24]:


ls ./uc/


# In[25]:


get_ipython().system('egrep BRILLOUIN ./uc/energy.out')


# ## PREPROCESS LEADS
# 
#     i) read in reversed k-points (convention uses +$k_x$)
#     ii) transpose matrix A[k]=A[-k].T

# In[26]:


pl_path


# In[27]:


h_pl_k_files = [pl_path/f'sc-KS_SPIN_1_K_{i}-1_0.csr' for i in range(48,0,-1)]
s_pl_k_files = [pl_path/f'sc-S_SPIN_1_K_{i}-1_0.csr' for i in range(48,0,-1)]


# In[28]:


read_bincsr = lambda file: read_bincsr_matrix(file).T.copy('C')
h_pl_k = np.stack(list(map(read_bincsr, h_pl_k_files)))
s_pl_k = np.stack(list(map(read_bincsr, s_pl_k_files)))


# In[29]:


plt.spy(h_pl_k[-1])


# In[30]:


atoms_pl = read(pl_path/'pl.xyz')
basis_pl = basis.Basis.from_dictionary(atoms_pl, {'C':13,'Si':13,'O':13})


# In[31]:


pl_path/'energy.out'


# In[32]:


get_ipython().system('egrep -i fermi /usr/scratch/mont-fort18/ggandus/k-point/uc/energy.out')


# In[33]:


fermi_pl = 0.33495567601659 * Hartree
fermi_pl


# In[34]:


h_pl_k *= Hartree
h_pl_k -= fermi_pl * s_pl_k


# In[35]:


kpts_t, h_pl_kii, s_pl_kii, h_pl_kij, s_pl_kij = tools.prepare_leads_matrices(basis_pl, h_pl_k, s_pl_k, (6,8,2))


# In[25]:


del h_pl_k
del s_pl_k


# In[26]:


kpts_t


# ### SAVE

# In[27]:


np.save('hs_pl_ii_ij',(h_pl_kii, s_pl_kii, h_pl_kij, s_pl_kij))


# In[36]:


plt.spy(h_pl_kij[0])


# # END OF PREPROCESSING

# In[67]:


get_ipython().system('hostname')


# In[68]:


get_ipython().system('pwd')


# In[ ]:


# %load trans.py
#!/usr/bin/env python

import numpy as np
from matplotlib import pyplot as plt


from qtpyt.abc import greenfunction, selfenergy
from qtpyt.surface import basis, principallayer, kpts


from ase.io import read


# In[ ]:


atoms_cc = read('sc.xyz')
basis_cc = basis.Basis.from_dictionary(atoms_cc, {'C':13,'Si':13,'O':13})

atoms_pl = read('pl.xyz')
basis_pl = basis.Basis.from_dictionary(atoms_pl, {'C':13,'Si':13,'O':13})

h_cc_kii, s_cc_kii, h_cc_kij, s_cc_kij = np.load('hs_cc_ii_ij.npy')
h_pl_kii, s_pl_kii, h_pl_kij, s_pl_kij = np.load('hs_pl_ii_ij.npy')


# In[50]:


kpts_t = kpts.monkhorst_pack((1,8,2))
Kpts = kpts.monkhorst_pack((1,4,2), offset=(0,0,0))
k_groups = kpts.unfold_kpts(Kpts, kpts_t)


# In[51]:


kpts_t


# In[52]:


Kpts


# In[54]:


k_groups = np.array(k_groups).reshape(-1,4)


# In[55]:


k_groups


# In[58]:


Kpts = kpts.monkhorst_pack((1,4,1), offset=(0,0,0))
Kpts


# In[49]:


Nr = (1,2,2) # # of realspace PLs in SC
Nk = (1,4,1) # # of SC transverse k-points


# In[ ]:


energies = np.linspace(-5,5,500,endpoint=True)
T_e = np.zeros_like(energies)


# In[ ]:


for i in range(np.prod(Nk)):

    selfenergies = [None, None]
    selfenergies[0] = principallayer.PrincipalSelfEnergy(kpts_t[k_groups[i]], (h_pl_kii[k_groups[i],...], s_pl_kii[k_groups[i],...]), (h_pl_kij[k_groups[i],...], s_pl_kij[k_groups[i],...]), Nr=Nr)
    selfenergies[1] = principallayer.PrincipalSelfEnergy(kpts_t[k_groups[i]], (h_pl_kii[k_groups[i],...], s_pl_kii[k_groups[i],...]), (h_pl_kij[k_groups[i],...], s_pl_kij[k_groups[i],...]), id='right', Nr=Nr)

    align_ao = basis_pl.nao
    diff = (h_cc_kii[i][align_ao, align_ao] - selfenergies[0].h_ii[align_ao, align_ao]) / s_cc_kii[i][align_ao, align_ao]
    h_cc_kii[i] -= diff * s_cc_kii[i]
    h_cc_kij[i] -= diff * s_cc_kij[i]

    internalselfenergies = [None, None]
    indices = np.ix_(range(selfenergies[0].nbf_i),range(selfenergies[0].nbf_i))
    internalselfenergies[0] = selfenergy.SelfEnergy((h_cc_kii[i], s_cc_kii[i]), (h_cc_kij[i], s_cc_kij[i]), [(indices, selfenergies[0])])
    internalselfenergies[1] = selfenergy.SelfEnergy((h_cc_kii[i], s_cc_kii[i]), (h_cc_kij[i], s_cc_kij[i]), [(indices, selfenergies[1])], id='right')

    indices = slice(None)
    gf = greenfunction.GreenFunction(internalselfenergies[0].H, internalselfenergies[0].S, [(slice(None), internalselfenergies[0]), (slice(None), internalselfenergies[1])])

    for e, energy in enumerate(energies):
        T_e[e] += gf.get_transmission(energy)
#         T_e[e] += gf.get_pdos(energy).sum(...some_indices...)

T_e /= np.prod(Nk)

np.save('ET', (energies, T_e))
plt.plot(energies, T_e)
plt.xlabel('E-E$_f$')
plt.ylabel('T(E)')
plt.savefig('ET.png',dpi=300)




import numpy as np
from collections import namedtuple

fields = ['M_a','nao_a','nao','atoms']
Calc = namedtuple('calc', fields)

def initialize_calculator(calc):
    calc.atoms.set_calculator(calc)
    if calc.wfs.S_qMM is None:
        # Initialize calculator
        calc.wfs.set_positions(calc.spos_ac)

def make_calc(atoms, basis):
    M_a = np.zeros(len(atoms),dtype=int)
    nao_a = np.zeros(len(atoms),dtype=int)
    nao = 0
    # Init
    M_a[0] = 0
    nao_a[0] = basis[atoms[0].symbol]
    nao += nao_a[0]
    for a0 in range(1,len(atoms)):
        nao_a[a0] = basis[atoms[a0].symbol]
        M_a[a0] = M_a[a0-1] + nao_a[a0-1]
        nao += nao_a[a0]
    return Calc(M_a, nao_a, nao, atoms)

def get_info(calc, attributes):
    info = []
    gpaw = hasattr(calc, 'wfs')
    # From gpaw calculator
    if gpaw:
        initialize_calculator(calc)
        natoms = len(calc.atoms)
        for attr in attributes:
            if attr in ['M_a','nao']:
                info.append(getattr(calc.setups, attr))
            elif attr in ['nao_a']:
                info.append([calc.setups[a0].nao for a0 in range(natoms)])
            else:
                raise NotImplementedError('{}'.format(attr))
    else:
        # From namedtuple
        for attr in attributes:
            info.append(getattr(calc, attr))
    return info

from .tk_gpaw import get_atom_indices, \
                     get_bfs_indices, \
                     get_bfs_atoms

class calcWRAP:

    def __init__(self, atoms, basis=None, default='dzp'):
        self.bfs_ai = get_bfs_atoms(atoms, basis=basis, default=default,
                            method='append')
        M_a, nao_a, nao = self.bfs2info(self.bfs_ai)
        self.nao_a = nao_a
        self.nao = nao
        self.M_a = M_a
        self.atoms = atoms
        #super().__init__(M_a, nao_a, nao, atoms)

    @staticmethod
    def bfs2info(bfs_ai):
        nao_a = np.array([len(bfs) for bfs in bfs_ai])
        M_a = [0] + \
              np.cumsum(nao_a)[:-1].tolist()
        M_a = np.array(M_a)
        nao = sum(nao_a)
        return M_a, nao_a, nao

    def get_bfs_indices(self, a, method='extend'):
        return get_bfs_indices(self, a, method=method)

    def get_atom_indices(self, a):  
        return get_atom_indices(self, a)



from qtpyt import xp
from qtpyt.abc._kernels import (_mulc,_mul,dottrace,dotdiag)

def test_mul():
    a = xp.random.random((10,10)).astype(complex)
    b = a.copy()
    c = a.copy()
    z = complex(2.)

    _mul(z,a,b,out=c)
    assert xp.allclose(z*b-a,c)


def test_mulc():
    a = xp.random.random((10,10)).astype(complex)
    b = a.copy()
    c = a.copy()
    z = complex(2.)

    _mulc(z,a.T,b.T,out=c)
    assert xp.allclose(z*b.T.conj()-a.T.conj(),c)


def test_dotx():
    a = xp.random.random((10,10)).astype(complex)
    b = a.copy()

    res = dottrace(a,b)
    assert xp.allclose(res, a.dot(b).trace())

    res = xp.zeros(a.shape[0],complex)
    dotdiag(a,b,res)
    assert xp.allclose(res, a.dot(b).diagonal())



from mpi4py import MPI
import numpy as np
import warnings

world = MPI.COMM_WORLD
rank = world.Get_rank()
size = world.Get_size()

MPI_type = {'int64':MPI.LONG,
            'float64':MPI.DOUBLE,
            'complex64':MPI.COMPLEX,
            'complex128':MPI.DOUBLE_COMPLEX,
            }

def print(obj, root=0, comm=world):
    if comm.Get_rank() == root:
        print(obj)

def get_stride(array, comm=world):
    n = array.size
    stride = n // comm.Get_size()
    if n % comm.Get_size() != 0:
        Warning('Cannot parallelize array.')
    return stride

def get_domain(array, comm=world):
    rank = comm.Get_rank()
    stride = get_stride(array, comm)
    out = np.empty(stride, dtype=array.dtype)
    out[:] = array[rank*stride:(rank+1)*stride]
    return out

def scatter(array, root=0, comm=world):
    stride = get_stride(array, comm)
    send_data = None
    send_count = 0
    send_type = MPI_type[str(array.dtype)]
    recv_count = stride
    recv_data = np.empty(recv_count, dtype=array.dtype)
    recv_type = send_type
    if comm.Get_rank() == root:
        send_data = array
        send_count = stride
    sendbuf = [send_data, send_count, send_type]
    recvbuf = [recv_data, recv_count, recv_type]
    comm.Scatter(sendbuf, recvbuf, root)
    return recv_data

def gather(array, root=0, comm=world):
    recv_data = None
    recv_count = 0
    recv_type = MPI_type[str(array.dtype)]
    send_data = array
    send_type = recv_type
    send_count = array.size
    if comm.Get_rank() == root:
        recv_count = send_count
        recv_data  = np.empty(recv_count * comm.Get_size(), dtype=array.dtype)
    sendbuf = [send_data, send_count, send_type]
    recvbuf = [recv_data, recv_count, recv_type]
    comm.Gather(sendbuf, recvbuf, root)
    return recv_data



import numpy as np
from collections import deque, namedtuple, defaultdict
from qtpyt.tk_gpaw import get_bfs_indices

cutoff = 1e-4

def get_graph(calc, h, s, cutoff=cutoff):

    num_vertices = len(calc.atoms)
    bfs_ai = get_bfs_indices(calc, range(num_vertices), 'append')

    # Vertex
    class vertex:
        '''
            :tag       level set index (block index in matrix)
            :neighbors neighborlist vertices (neighborlist atoms)
            :index     vertex in graph (atom index in atoms)
        '''
        def __init__(self, tag, neighbors):
            self.tag = tag
            self.neighbors = neighbors
            self.index = 0
    # Graph ::dict of vertices
    graph  = defaultdict(lambda : vertex(0,[]))

    # Start
    for a0 in range(num_vertices):
        # Find neighborlist for a1 > a0
        for a1 in range(a0,len(calc.atoms)):
            h_ij = h.take(bfs_ai[a0],axis=0).take(bfs_ai[a1],axis=1)
            if np.any(h_ij>cutoff):
                graph[a0].neighbors.append(a1)
                graph[a0].index = a0
                # Do not double count a0
                if a0 != a1:
                    # insert neighbor a0 in a1.neighborlist
                    graph[a1].neighbors.append(a0)

    return graph

def BFS(Q, graph, subset):
    '''
        :Q      deque with first vertices
        :graph  collections::defaultdict of ::vertex
        :subset restrict search to subset of vertices
    '''

    subset = list(subset)
    # Start
    while Q:
        v = Q.pop()
        adjacentEdges = graph[v].neighbors #grapht_neighbors(v)[0]
        #
        for w in adjacentEdges:
            # Atom not discovered and in subset
            if (not graph[w].tag) & (w in subset):
                # Assign parent previous level set
                graph[w].tag = graph[v].tag + 1 if graph[v].tag > 0 else graph[v].tag - 1
                Q.appendleft(w)


def get_tridiagonal_nodes(calc, h, s, pl1, pl2=None, cutoff=cutoff):
    '''
        :pl1 number of atoms in left principal layer
        :pl2 number of atoms in right principal layer
    '''

    if pl2 is None:
        pl2 = pl1

    # Construct graph
    graph = get_graph(calc, h, s, cutoff)

    # Number of vertices
    num_vertices = len(graph)

    # Pinn first and last l.s.
    v0 = range(pl1)
    vN = range(num_vertices-pl2,num_vertices)
    _set_tags(graph, v0, 1)
    _set_tags(graph, vN, -1)

    # BFS from left for left subset vertices
    Q  = deque(range(pl1))
    subset = range(num_vertices//2)
    BFS(Q, graph, subset)
    # BFS from right for right subset vertices
    Q  = deque(range(num_vertices-1,num_vertices-pl2-1,-1))
    subset = range(num_vertices//2,num_vertices)
    BFS(Q, graph, subset)

    # from IPython import embed
    # embed()

    # Assign last l.s. from right to last l.s. from left
    tags = _get_tags(graph)
    tmin = min(tags)
    tmax = max(tags)
    for a in graph.values():
        if a.tag == tmin: a.tag = tmax

    # Rename positive
    tags = _get_tags(graph)
    u = np.unique(tags).tolist()
    for a in graph.values():
        # Start from 0 index
        if a.tag > 0:
            a.tag -= 1
        # Map negative to positive
        else:
            a.tag %= len(u)

    # Get nodes:= cut edges
    tags = _get_tags(graph)
    u = np.unique(tags).tolist()
    nodes = [0] # N+1 nodes
    for u0 in u:
        indices = _get_indices(graph, u0)
        bfs_i = get_bfs_indices(calc, int(np.max(indices)))
        nodes.append(np.max(bfs_i)+1)

    return nodes

def split_tridiagonal(mat, nodes):


    N = len(nodes) - 1

    mat_list_ii = [None for _ in range(N)]
    mat_list_ij = [None for _ in range(N-1)]

    for q in range(N):
        q0 = nodes[q]
        q1 = nodes[q+1]
        row = mat.take(range(q0, q1), axis=0)
        mat_list_ii[q] = row.take(range(q0, q1), axis=1)
        if q < (N-1):
            q2 = nodes[q+2]
            mat_list_ij[q] = row.take(range(q1, q2), axis=1)

    return mat_list_ii, mat_list_ij

def tridiagonalize(calc, h, s, pl1, pl2=None, cutoff=cutoff):

    nodes = get_tridiagonal_nodes(calc, h, s, pl1, pl2, cutoff)

    h_list_ii, h_list_ij = split_tridiagonal(h, nodes)
    s_list_ii, s_list_ij = split_tridiagonal(s, nodes)

    return (h_list_ii, s_list_ii), (h_list_ij, s_list_ij)

################# Helpers ###################

def _set_tags(graph, indices, tag):
    if isinstance(indices, int):
        indices = [indices]
    for i in indices:
        graph[i].tag = tag

def _get_tags(graph):
    return [a.tag for a in graph.values()]

def _get_indices(graph, tag):
    if isinstance(tag, int):
        tag = [tag]
    return sorted([a.index for a in graph.values() if a.tag in tag])



import numpy as np
from scipy import linalg as la
from ase.calculators.singlepoint import SinglePointCalculator
from gpaw.setup import types2atomtypes
from gpaw.basis_data import Basis
from .tools import subdiagonalize, get_orthonormal_subspace, \
                   get_subspace, rotate_matrix, dagger
from .block import block

def initialize_calculator(calc):
    calc.atoms.set_calculator(calc)
    if calc.wfs.S_qMM is None:
        # Initialize calculator
        calc.wfs.set_positions(calc.spos_ac)

def get_atom_indices(calc, a):
    from .tk_calc import get_info
    M_a, nao_a, nao = get_info(calc, ['M_a','nao_a','nao'])
    if isinstance(a, (int,str)):
        a = [a]
    ai = []
    for elem in a:
        if isinstance(elem, int):
            ai.append(elem)
        elif isinstance(elem, str):
            symbol = elem
            for a0 in range(len(calc.atoms)):
                offset = 0
                if a0 >= len(M_a): # original number of atoms,
                                               # e.g. allows calc.atoms = atoms.repeat(..)
                    offset += (a0 // len(M_a)) * len(M_a)
                    a0 %= len(M_a)
                if calc.atoms[a0].symbol==symbol:
                    ai.append(a0+offset)
        else:
            raise Warning('Non valid input {}'.format(elem))
    ai = sorted(ai)
    return ai

def get_bfs_indices(calc, a, method='extend'):
    """Get list of basis function indices of atom(s) a.
    Available methods:
        1. 'extend' - return flatten ao indices
        2. 'append' - return ao.shape(natoms,nao)
    """
    from .tk_calc import get_info
    M_a, nao_a, nao = get_info(calc, ['M_a','nao_a','nao'])
    if isinstance(a, (int,np.int32,np.int64)):
        a = [a]
    Mvalues = []
    Mattr = getattr(Mvalues, method)
    for a0 in a:
        M = 0
        M += M_a[a0]
        Mattr(list(range(M, M + nao_a[a0])))
    return Mvalues

def get_bfs_atoms(atoms, basis=None, default='dzp', method='extend'):
    """Get ba
    atoms - ase::atoms
    basis - dict of elements with correspondig basis set type
    default - str
    """
    if basis is None:
        basis = {symbol:default for symbol in set(atoms.symbols)}
    symbols = atoms.get_chemical_symbols()
    basis_a = types2atomtypes(symbols, basis, default)

    nao_a = [Basis(symbol, type).nao
             for symbol, type in zip(symbols, basis_a)]

    Mvalues = []
    Mattr = getattr(Mvalues, method)
    i0 = 0
    for nao in nao_a:
        i1 = i0 + nao
        Mattr(list(range(i0,i1)))
        i0 = i1
    return Mvalues

def subdiagonalize_atoms(calc, h_ii, s_ii, a=None):
    """Get rotation matrix for subdiagonalization of given(all) atoms."""
    from scipy.linalg import block_diag
    if a is None:
        a = range(len(calc.atoms))
    if isinstance(a, int):
        a = [a]
    U_mm = []
    e_aj = []
    na_tot = len(calc.atoms)
    for a0 in range(na_tot):
        bfs = get_bfs_indices(calc, a0)
        if a0 in a:
            v_jj, e_j = get_orthonormal_subspace(h_ii,s_ii,bfs)
            U_mm.append(v_jj)
            e_aj.append(e_j)
        else:
            U_mm.append(np.eye(len(bfs)))
    U_mm = block_diag(*U_mm)
    return U_mm, e_aj

def subdiagonalize_blocks(calc, h_ii, s_ii, a=None):
    """Get rotation matrix for subdiagonalization of given(all) atoms."""
    from scipy.linalg import block_diag
    U_mm = []
    e_aj = []
    for a0 in a:
        bfs = get_bfs_indices(calc, a0)
        v_jj, e_j = get_orthonormal_subspace(h_ii,s_ii,bfs)
        U_mm.append(v_jj)
        e_aj.append(e_j)
    U_mm = block_diag(*U_mm)
    return U_mm, e_aj

def get_bf_centers(atoms, basis=None):

    from ase.calculators.singlepoint import SinglePointCalculator
    from gpaw.setup import types2atomtypes

    calc = atoms.get_calculator()
    if calc is None or isinstance(calc, SinglePointCalculator):
        symbols = atoms.get_chemical_symbols()
        basis_a = types2atomtypes(symbols, basis, 'dzp')
        nao_a = [Basis(symbol, type).nao
                 for symbol, type in zip(symbols, basis_a)]
    else:
        if not calc.initialized:
            calc.initialize(atoms)
        nao_a = [calc.wfs.setups[a].nao for a in range(len(atoms))]
    pos_ic = []
    for pos, nao in zip(atoms.get_positions(), nao_a):
        pos_ic.extend(pos[None].repeat(nao, 0))
    return np.array(pos_ic)

def sum_bf_atom(calc, X_i, n_a=0):
    '''Sum basis functions on the same atom.'''
    if n_a == 0:
        n_a = len(calc.atoms)
    X_a = np.zeros(n_a, dtype=X_i.dtype)
    i0 = 0
    for a0 in range(n_a):
        i1 = i0 + calc.wfs.setups[a0].nao
        X_a[a0] = sum(X_i[i0:i1])
        i0 = i1
    return X_a

def flatten(iterables):
    return (elem for iterable in iterables for elem in iterable)

def get_U1(bfs_m, bfs_i, c_MM=None, apply=False):
    nbf=len(bfs_m) + len(bfs_i)
    if c_MM is None:
        c_MM = np.eye(nbf)
    U_mm = c_MM.take(bfs_m,axis=1)
    U_ii = c_MM.take(bfs_i,axis=1)
    U1 = np.array(block([[1,0],[U_ii,U_mm]]))
    """If apply is True permute inplace bfs_m bfs_i"""
    if apply:
        bfs_m[:] = range(nbf-len(bfs_m), nbf)
        bfs_i[:] = range(nbf-len(bfs_m))
    return U1

def get_U2(s1_MM, bfs_m, bfs_i):
    s_mm = get_subspace(s1_MM, bfs_m)
    s_mi = s1_MM.take(bfs_m,axis=0).take(bfs_i,axis=1)
    U_mi = - la.inv(s_mm).dot(s_mi)
    U_ii = np.eye(len(bfs_i))
    U2 = np.array(block([[U_ii,0],[U_mi,np.eye(len(bfs_m))]]))
    return U2

def get_mm_ii_im(h_MM, s_MM, bfs_m, bfs_i):
    h_ii = get_subspace(h_MM, bfs_i)
    s_ii = get_subspace(s_MM, bfs_i)
    h_mm = get_subspace(h_MM, bfs_m)
    s_mm = get_subspace(s_MM, bfs_m)
    h_im = h_MM.take(bfs_i,axis=0).take(bfs_m,axis=1)
    s_im = s_MM.take(bfs_i,axis=0).take(bfs_m,axis=1)
    return (h_mm, s_mm), (h_ii, s_ii), (h_im, s_im)

def extract_orthogonal_subspaces(h_MM, s_MM, bfs_m, bfs_i, c_MM=None, orthogonal=True, apply=True):
    U1 = get_U1(bfs_m, bfs_i, c_MM, apply=True) # Modifies bfs_m, bfs_i
    h1_MM = rotate_matrix(h_MM,U1)
    s1_MM = rotate_matrix(s_MM,U1)
    """If orthogonal is True set s_im to zeros."""
    if orthogonal:
        U2 = get_U2(s1_MM, bfs_m, bfs_i)
        h2_MM = rotate_matrix(h_MM, U1.dot(U2))
        s2_MM = rotate_matrix(s_MM, U1.dot(U2))
    else:
        h2_MM = h1_MM
        s2_MM = s1_MM
    if apply is False:
        return U1.dot(U2)
    return get_mm_ii_im(h2_MM, s2_MM, bfs_m, bfs_i)


def get_orthogonal_subspaces(calc, h_MM, s_MM, a=None, cutoff=np.inf, orthogonal=True):
    """Split matrix in hybridization and active spaces.
       Active space contains orbital(s) of atom(s) a with
       abs(energy) < cutoff"""
    nbf=len(h_MM)
    c_MM, e_aj = subdiagonalize_atoms(calc, h_MM, s_MM, a)
    bfs_imp = get_bfs_indices(calc, a)
    bfs_m = [bfs_imp[i] for i, eigval in enumerate(flatten(e_aj)) if abs(eigval)<cutoff]
    bfs_i = list(np.setdiff1d(range(nbf),bfs_m))
    return extract_orthogonal_subspaces(h_MM, s_MM, bfs_m, bfs_i, c_MM, orthogonal)

def get_ll_mm_rr(H, S, nbf_l, nbf_m, nbf_r):
    rng_l = list(range(nbf_l))
    rng_m = list(range(nbf_l,nbf_l+nbf_m))
    rng_r = list(range(nbf_l+nbf_m,nbf_l+nbf_m+nbf_r))
    s_lm = S.take(rng_l,0).take(rng_m,1)
    h_lm = H.take(rng_l,0).take(rng_m,1)
    s_rm = S.take(rng_r,0).take(rng_m,1)
    h_rm = H.take(rng_r,0).take(rng_m,1)
    h_mm = get_subspace(H, rng_m)
    s_mm = get_subspace(S, rng_m)
    return (h_lm, s_lm), (h_mm, s_mm), (h_rm, s_rm)

def get_lead_orthogonal(greenfunction):
    s_ll = greenfunction.selfenergies[0].s_ii
    s_rr = greenfunction.selfenergies[1].s_ii
    h_ll = greenfunction.selfenergies[0].h_ii
    h_rr = greenfunction.selfenergies[1].h_ii
    s_lm = greenfunction.selfenergies[0].s_im
    s_rm = greenfunction.selfenergies[1].s_im
    h_lm = greenfunction.selfenergies[0].h_im
    h_rm = greenfunction.selfenergies[1].h_im
    h_mm = greenfunction.H
    s_mm = greenfunction.S
    nbf_l = len(h_ll)
    nbf_r = len(h_rr)
    nbf_m = len(h_mm)
    s_ml = dagger(s_lm)
    s_mr = dagger(s_rm)
    h_ml = dagger(h_lm)
    h_mr = dagger(h_rm)
    U_lm = - la.inv(s_ll).dot(s_lm)
    U_rm = - la.inv(s_rr).dot(s_rm)
    U = np.array(block([[np.eye(nbf_l),U_lm,0],[0,np.eye(nbf_m),0],[0,U_rm,np.eye(nbf_r)]]))
    S = np.array(block([[s_ll,s_lm,0],[s_ml,s_mm,s_mr],[0,s_rm,s_rr]]))
    H = np.array(block([[h_ll,h_lm,0],[h_ml,h_mm,h_mr],[0,h_rm,h_rr]]))
    S = rotate_matrix(S,U)
    H = rotate_matrix(H,U)
    return get_ll_mm_rr(H, S, nbf_l, nbf_m, nbf_r)



import numpy as np
import scipy.linalg as la
# Physical quantities
from scipy.constants import e, k


'''References:

http://www1.spms.ntu.edu.sg/~ydchong/teaching/08_contour_integration.pdf
'''

def zero_fermi(nzp):
    '''Compute poles (zp) and residues (Rp) of fermi function.'''
    N = nzp
    M = 2*N

    A = np.zeros((2*M,2*M))
    B = np.zeros((2*M,2*M))

    zp = np.zeros(2+M)
    Rp = np.zeros(2+M)

    for i in range(1,M+1):
        B[i,i] = (2*i-1)

    for i in range(1,M):
        A[i,i+1] = -0.5;
        A[i+1,i] = -0.5;

    a = np.zeros(M*M)
    b = np.zeros(M*M)

    for i in range(M):
        for j in range(M):
            a[j*M+i] = A[i+1,j+1]
            b[j*M+i] = B[i+1,j+1]

    a.shape = (M,M)
    b.shape = (M,M)

    eigvas, eigvecs = la.eigh(a,b)

    zp[:M] = eigvas

    for i in range(M,0,-1):
        zp[i] = zp[i-1]

    for i in range(1,M+1):
        zp[i] = 1.0/zp[i];

    a = eigvecs.T.flatten()

    for i in range(0,M):
        Rp[i+1] = -a[i*M]*a[i*M]*zp[i+1]*zp[i+1]*0.250;

    zp = -zp[1:N+1]
    Rp = Rp[1:N+1]

    return zp, Rp

def integrate_dos(G, mu=0, T=300, nzp=100, R=1e10):

    zp, Rp = zero_fermi(nzp)
    N = nzp

    k_B = k / e # Boltzmann constant [eV/K] 8.6173303e-05
    beta = 1/(k_B*T)
    a_p = mu + 1j*zp/beta

    eta = G.eta
    G.eta = complex(0.)

    R = 1e10
    mu_0 = 1j*R*G.apply_overlap(1j*R, trace=True)

    mu_1 = complex(0)
    for i in range(N):
        mu_1 += G.apply_overlap(a_p[i], trace=True) * Rp[i]
    mu_1 *= -1j*4/beta

    rho = np.real(mu_0) + np.imag(mu_1)

    G.eta = eta

    return rho

def integrate_pdos(G, mu=0, T=300, nzp=100, R=1e10):

    zp, Rp = zero_fermi(nzp)
    N = nzp

    k_B = k / e # Boltzmann constant [eV/K] 8.6173303e-05
    beta = 1/(k_B*T)
    a_p = mu + 1j*zp/beta

    eta = G.eta
    G.eta = complex(0.)

    R = 1e10
    #SGS = G.S @ G.apply_overlap(1j*R)
    mu_0 = 1j*R * G.apply_overlap(1j*R, diag=True) #SGS.diagonal() / G.S.diagonal()

    mu_1 = np.zeros(len(mu_0),complex)
    for i in range(N):
        #SGS = G.S @ G.apply_overlap(a_p[i])
        mu_1 += G.apply_overlap(a_p[i], diag=True) * Rp[i] #SGS.diagonal() / G.S.diagonal() * Rp[i]
    mu_1 *= -1j*4/beta

    rho = np.real(mu_0) + np.imag(mu_1)

    G.eta = eta

    return rho



from qtpyt import xp
from qtpyt.abc._kernels import (_mul,_mulc,get_lambda,dottrace,dagger)


class GreenFunction():
    """Equilibrium retarded Green function."""

    def __init__(self, H, S=None, selfenergies=[], eta=1e-5):
        self.H = xp.asarray(H)
        self.S = xp.asarray(S)
        self.selfenergies = selfenergies

        self.eta = eta
        self.energy = None

        self.Ginv = xp.empty(H.shape, complex)
        self.gammas = [None for _ in range(len(selfenergies))]

    def get_Ginv(self, energy):
        if energy != self.energy:
            self.energy = energy
            z = energy + self.eta * 1.j
            self.Ginv[:] = _mul(z,self.H,self.S)
            # Add selfenergies
            for i, (indices, selfenergy) in enumerate(self.selfenergies):
                sigma = xp.asarray(selfenergy.retarded(energy))
                self.Ginv[indices] -= sigma
                self.gammas[i] = get_lambda(sigma)
 
        return self.Ginv

    def retarded(self, energy):
        G = self.get_Ginv(energy)
        return xp.linalg.inv(G)

    def apply_retarded(self, energy, X):
        """Apply retarded Green function to X.

        Returns the matrix product G^r(e) . X
        """
        Ginv = self.get_Ginv(energy)
        return xp.linalg.solve(Ginv, X)

    def get_dos(self, energy):
        """Total density of states -1/pi Im(Tr(GS))"""
        GS = self.apply_retarded(energy, self.S)
        return -GS.imag.trace() / xp.pi

    def get_pdos(self, energy):
        """Total density of states -1/pi Im(Tr(GS))"""
        GS = self.apply_retarded(energy, self.S)
        return -GS.imag.diagonal() / xp.pi

    def get_transmission(self, energy):

        Ginv = self.get_Ginv(energy)
        gamma_L = self.gammas[0]
        gamma_R = self.gammas[1]
        # TODO :: improve with single factorization compatible with GPU
        a_mm = xp.linalg.solve(Ginv, gamma_L)
        b_mm = xp.linalg.solve(dagger(Ginv), gamma_R)
        T_e = dottrace(a_mm, b_mm).real

        return T_e



import argparse
import sys
import os
import re
from pathlib import Path

import numpy as np

from ase.io import read

from qttools.cp2k.tools import (read_basis_dict, read_fermi, read_mp_kmesh, read_reversed_from_folder)
from qtpyt.surface.basis import Basis
from qtpyt.surface.tools import prepare_leads_matrices


class ParseKwargs(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, dict())
        for value in values:
            key, value = value.split('=')
            getattr(namespace, self.dest)[key] = value


def main():
    parser = argparse.ArgumentParser(argument_default=None)
    parser.add_argument('--folder',
                        help='Working directory',
                        type=str,
                        default=os.getcwd(),
                        required=False)
    parser.add_argument('--out','-o',
                        help='atomic structure file',
                        type=str,
                        default=r'ener[gy]*\.o[ut]*',
                        required=False)
    parser.add_argument('--atoms',
                        help='atomic structure file',
                        type=str,
                        default=r'.*\.xyz',
                        required=False)
    parser.add_argument('--fermi',
                        type=float,
                        help='Fermi level in eV',
                        required=False)
    parser.add_argument('--kmesh',
                        type=int,
                        nargs=3,
                        help='Monkhorst-Pack meshgird',
                        required=False)
    parser.add_argument('--basis',
                        nargs='*',
                        help='Basis dictionary',
                        required=False,
                        default=None,
                        action=ParseKwargs)
    parser.add_argument('--offset',
                        type=int,
                        nargs=3,
                        default=(0.,0.,0.),
                        help='Monkhorst-Pack offset',
                        required=False)

    argv = parser.parse_args()

    path = Path(argv.folder)
    files = [p.name for p in path.glob('*') if p.is_file()]

    search_atoms = re.compile(argv.atoms)
    search_out = re.compile(argv.out)

    atoms = list(filter(search_atoms.match, files))
    out = list(filter(search_out.match, files))

    # _varname = lambda var: [ k for k,v in globals().items() if v == var][0]
    # _varname.__doc__ = "Get variable name as string."

    def _check(*args):
        """Check that parsed argument files are single and exist."""
        for arg, name in args:
            # name = _varname(arg)
            if not arg:
                raise RuntimeError(f"""
                    File not found {name} {arg}. Use --{name} at input.""")
            if len(arg)>1:
                raise RuntimeError(f"""
                    Found multiple matches for {name} {arg}. Use --{name} at input.""")

    _check((out,'out'), (atoms,'atoms'))
    out_file = out[0]
    atoms_file = atoms[0]

    atoms = read(atoms_file)
    basis_dict = argv.basis or read_basis_dict(out_file)
    basis = Basis.from_dictionary(atoms, basis_dict)
    fermi = argv.fermi or read_fermi(out_file)
    kmesh = argv.kmesh or read_mp_kmesh(out_file)
    offset = argv.offset

    print(f"""Input params:
        atoms : {atoms_file},
        out : {out_file},
        basis : {basis_dict},
        kmesh : {kmesh},
        offset : {offset},
        fermi : {fermi}""")

    h_k, s_k = read_reversed_from_folder(path)
    h_k -= fermi * s_k

    kpts_t, h_kii, s_kii, h_kij, s_kij = prepare_leads_matrices(basis, h_k, s_k, kmesh, offset)

    del h_k
    del s_k

    np.save(path/'hs_kii_kij', (h_kii, s_kii, h_kij, s_kij))
    np.savetxt(path/'kpts.txt', kpts_t, fmt='%6.3f')

if __name__ == '__main__':
    sys.exit(main())


import numpy as np
import numbers

class Basis:
    """Basis function descriptor.
    
    Example :
    In [1]: from ase import Atoms
    In [2]: from qtpyt.surface.basis import Basis
    In [2]: 
    In [2]: atoms = Atoms('OH',positions=[[0,0,0],[0.5,0.5,0.5]],cell=[1,1,1])
    In [3]: basis = Basis.from_dictionary(atoms, {'O':2,'H':1})
        
    """

    def __init__(self, atoms, nao_a, M_a=None):
        """Build descriptor from atoms and # of atomic orbitals
        per atom nao.

        """
        self.atoms = atoms
        self.nao_a = np.array(nao_a, ndmin=1)
        if M_a is None:
            M_a = np.cumsum(np.insert(self.nao_a[:-1], 0, 0))
        self.M_a = np.array(M_a, ndmin=1)
        self.nao = np.sum(self.nao_a) # Total # of atomic orbitals

    @classmethod
    def from_dictionary(cls, atoms, basis):
        """Build descriptor from atoms and dictionary of key value
        paris {symbol : nao}.

        """
        nao_a = [basis[symbol] for symbol in atoms.symbols]
        return cls(atoms, nao_a)

    @property
    def centers(self):
        """Expand atomic centers to basis functions.
        
        Example:
        In [1]: basis.atoms.positions
        Out[1]: 
        array([[0. , 0. , 0. ],
            [0.5, 0.5, 0.5]])

        In [2]:
        In [2]: basis.centers
        Out[2]: 
        array([[0. , 0. , 0. ],
            [0. , 0. , 0. ],
            [0.5, 0.5, 0.5]])


        """
        return self._expand(self.atoms.positions)

    def repeat(self, N):
        """Create a new repeated basis object.

        Args:
            N : sequence of three positives.

        Example:
        In [1]: basis.nao_a
        Out[1]: array([2, 1])

        In [2]: basis.repeat((1,2,1)).nao_a
        Out[2]: array([2, 1, 2, 1])

        """
        nao_a = np.tile(self.nao_a, np.prod(N))
        atoms = self.atoms.repeat(N)
        return Basis(atoms, nao_a)

    def argsort(self, order='xyz'):
        """Sort basis indices using corresponding atomic positions.
        
        The first character is used as primary coordinate and so on.
        NOTE : Opposite to np.lexsort

        Example:
        In [2]: basis.nao_a
        Out[2]: array([2, 1, 2, 1])

        In [38]: basis.atoms.positions
        Out[38]: 
        array([[0. , 0. , 0. ],
            [0.5, 0.5, 0.5],
            [0. , 1. , 0. ],
            [0.5, 1.5, 0.5]])

        In [39]: basis.atoms.positions[basis.argsort()]
        Out[39]: 
        array([[0. , 0. , 0. ],
            [0. , 1. , 0. ],
            [0.5, 0.5, 0.5],
            [0.5, 1.5, 0.5]])

        In [41]: basis.nao_a[basis.argsort()]
        Out[41]: array([2, 2, 1, 1])

        """
        positions = self.atoms.positions
        i = order.index('x')
        j = order.index('y')
        k = order.index('z')
        idx = np.lexsort((positions[:,k],positions[:,j],positions[:,i]))
        return idx

    def get_indices(self, indices=None):
        """Get basis function indices.
        
        Args:
            indices : (optional) take subset/sorted atomic indices.
        """
        if indices is None:
            M_a, nao_a, nao = self.M_a, self.nao_a, self.nao
        elif isinstance(indices, numbers.Integral):
            return np.arange(self.nao_a[indices])+self.M_a[indices]
        else:
            M_a = self.M_a[indices]
            nao_a = self.nao_a[indices]
            nao = np.sum(nao_a)
            
        idx = self._expand(M_a, nao_a)
        i0 = M_a[0]
        for i in range(1,nao):
            if idx[i] == i0:
                idx[i] = idx[i-1] + 1
            else:
                i0 = idx[i]
        return idx

    def __getitem__(self, i):
        """Return a subset of the atoms."""
        return self.__class__(self.atoms[i], self.nao_a[i], self.M_a[i])

    def __len__(self):
        return self.nao

    def _expand(self, array, repeats=None):
        """Helper function. Expand atoms array to basis functions."""
        if repeats is None:
            repeats = self.nao_a
        return np.repeat(array, repeats, axis=0)


import numpy as np
from numpy import linalg
from scipy.sparse import csr_matrix
# from functools import lru_cache
from .coupledhamiltonian import CoupledHamiltonian
from .tools import dagger

#Recursive GreenFunction helpers
from .solvers.tridiagonal   import tridiagonalize, cutoff
from .solvers.recursive import * #get_mat_lists, recursive_gf, multiply

#Density integral
from .continued_fraction import integrate_pdos
from .tk_gpaw import sum_bf_atom

# GPU
# import cupy as xp
from transport import xp

class GreenFunction(CoupledHamiltonian):
    """Equilibrium retarded Green function."""

    def __init__(self, H, S=None, selfenergies=[], **kwargs):
        self.H = H
        self.S = S
        self.selfenergies = selfenergies
        # self.eta = eta
        self.energy = None
        self.Ginv = np.empty(H.shape, complex)

        self.parameters = {'eta': 1e-5,
                           'align_bf': None}
        self.initialized = False
        self.set(**kwargs)

    def set(self, **kwargs):
        for key in kwargs:
            if key in ['eta','align_bf']:
                self.initialized = False
                break
            elif key not in self.parameters:
                raise KeyError('%r not a vaild keyword' % key)

        self.parameters.update(kwargs)

    def initialize(self):#, calc, pl1, pl2=None, cutoff=cutoff, align_bf=None):
        '''
            calc: scattering calculator
            pl1: # of bfs left principal layer
            pl2: # of bfs right principal layer (default to pl1)
            cutoff: cutoff for block neighbor interaction
        '''

        if self.initialized:
            return

        p = self.parameters
        align_bf = p['align_bf']

        #Set eta
        self.eta = p['eta']

        #Align first basis funciton for Left lead
        if align_bf is not None:
            self.align_bf(align_bf)

        self.initialized = True


    # @lru_cache(maxsize=None)
    def retarded(self, energy, inverse=False):
        """Get retarded Green function at specified energy.

        If 'inverse' is True, the inverse Green function is returned (faster).
        """
        if energy != self.energy:
            self.energy = energy
            z = energy + self.eta * 1.j

            if self.S is None:
                self.Ginv[:] = 0.0
                self.Ginv.flat[:: len(self.S) + 1] = z
            else:
                self.Ginv[:] = z
                self.Ginv *= self.S
            self.Ginv -= self.H

            for selfenergy in self.selfenergies:
                self.Ginv -= selfenergy.retarded(energy)

        if inverse:
            return self.Ginv
        else:
            return np.linalg.inv(self.Ginv)

    def calculate(self, energy, sigma):
        """XXX is this really needed"""
        ginv = energy * self.S - self.H - sigma
        return np.linalg.inv(ginv)

    def apply_retarded(self, energy, X):
        """Apply retarded Green function to X.

        Returns the matrix product G^r(e) . X
        """
        return np.linalg.solve(self.retarded(energy, inverse=True), X)

    def apply_overlap(self, energy, trace=False, diag=False):
        """Apply retarded Green function to S."""
        GS = self.apply_retarded(energy, self.S)
        if trace:
            return np.trace(GS)
        if diag:
            return GS.diagonal()
        return GS

    def dos(self, energy):
        """Total density of states -1/pi Im(Tr(GS))"""
        if self.S is None:
            return -self.retarded(energy).imag.trace() / np.pi
        else:
            GS = self.apply_retarded(energy, self.S)
            return -GS.imag.trace() / np.pi

    def pdos(self, energy):
        """Projected density of states -1/pi Im(SGS/S)"""
        if self.S is None:
            return -self.retarded(energy).imag.diagonal() / np.pi
        else:
            S = self.S
            SGS = np.dot(S, self.apply_retarded(energy, S))
            return -(SGS.diagonal() / S.diagonal()).imag / np.pi

    def take_bfs(self, bfs, apply):
        h_pp, s_pp, c_mm = CoupledHamiltonian.take_bfs(self, bfs, apply)
        if apply:
            self.Ginv = np.empty(self.H.shape, complex)
        return h_pp, s_pp, c_mm

    def get_spectrals(self, energy):
        """Specral density of states incoming from leads."""
        G_mm = self.retarded(energy)
        lambda1_mm = self.selfenergies[0].get_lambda(energy)
        lambda2_mm = self.selfenergies[1].get_lambda(energy)
        A1_mm = G_mm.dot(lambda1_mm).dot(dagger(G_mm))
        A2_mm = G_mm.dot(lambda2_mm).dot(dagger(G_mm))

        return A1_mm, A2_mm

    def get_transmission(self, energies, T_e=None):

        if T_e is None:
            T_e = np.zeros(len(energies))

        for e, energy in enumerate(energies):
            Ginv_mm = self.retarded(energy, inverse=True)
            lambda1_mm = self.selfenergies[0].get_lambda(energy)
            lambda2_mm = self.selfenergies[1].get_lambda(energy)
            a_mm = linalg.solve(Ginv_mm, lambda1_mm)
            b_mm = linalg.solve(dagger(Ginv_mm), lambda2_mm)
            T_e[e] = np.trace(np.dot(a_mm, b_mm)).real

        return T_e


class RecursiveGF(CoupledHamiltonian):

    def __init__(self, H=None, S=None, selfenergies=[], **kwargs):

        super().__init__(H, S, selfenergies)

        self.energy = None
        self.g1N = None

        self.parameters = {'eta': 1e-5,
                           'calc': None,
                           'pl1': None,
                           'pl2': None,
                           'cutoff': cutoff,
                           'align_bf': None,
                           'H': None,
                           'S': None,
                           'gpu': False}

        self.initialized = False
        self.set(**kwargs)

    def set(self, **kwargs):
        for key in kwargs:
            if key in ['eta','calc', 'pl1', 'pl2',
                       'cutoff', 'align_bf','H','S','gpu']:
                self.initialized = False
                break
            elif key not in self.parameters:
                raise KeyError('%r not a vaild keyword' % key)

        self.parameters.update(kwargs)

    def initialize(self, hs_list_ii=None, hs_list_ij=None):#, calc, pl1, pl2=None, cutoff=cutoff, align_bf=None):
        '''
            calc: scattering calculator
            pl1: # of bfs left principal layer
            pl2: # of bfs right principal layer (default to pl1)
            cutoff: cutoff for block neighbor interaction
        '''
        if self.initialized:
            return

        p = self.parameters
        align_bf = p['align_bf']
        calc = p['calc']
        pl1 = p['pl1']
        pl2 = p['pl2']
        cutoff = p['cutoff']
        H = p['H']
        S = p['S']

        if pl1 is None:
            pl1 = self.selfenergies[0].natoms
        if pl2 is None:
            pl2 = pl1

        #Set eta
        self.eta = p['eta']

        if hs_list_ii is None:
            #Construct block tridiagonal lists
            hs_list_ii, hs_list_ij = tridiagonalize(
                                               calc, H, S,
                                               pl1, pl2, cutoff)
        # Offload to GPU
        if p['gpu'] is True:
            hs_list_ii = [[xp.asarray(h) for h in hs_list_ii[0]],
                   [xp.asarray(s) for s in hs_list_ii[1]]]
            hs_list_ij = [[xp.asarray(h) for h in hs_list_ij[0]],
                   [xp.asarray(s) for s in hs_list_ij[1]]]
        else:
            from .solvers import recursive
            recursive.xp = np
        # else:
        self.hs_list_ii, self.hs_list_ij = list(hs_list_ii), list(hs_list_ij)

        # self.truncate(hs_list_ii)
        # self.truncate(hs_list_ij)

        #Align first basis funciton for Left lead
        if align_bf is not None:
            self.align_bf(align_bf)

        self.hs_list_ji = [[h.T for h in self.hs_list_ij[0]],
                           [s.T for s in self.hs_list_ij[1]]]

        # self.dense2csr(self.hs_list_ii)
        # self.dense2csr(self.hs_list_ij)
        # self.dense2csr(self.hs_list_ji)

        #Note h_im[:,:pl1]=h_ij for Left and h_im[:,-pl2:]=h_ij for Right
        #for selfenergy in self.selfenergies:
        #    selfenergy.h_im = selfenergy.h_ij
        #    selfenergy.s_im = selfenergy.s_ij
        #    selfenergy.sigma_mm = np.empty((selfenergy.nbf_i, selfenergy.nbf_i),
        #                                    dtype=selfenergy.sigma_mm.dtype)

        self.nbf = sum(h.shape[0] for h in self.hs_list_ii[0])
        self.N = len(self.hs_list_ii[0])
        self.initialized = True

    @staticmethod
    def truncate(hs_list, cutoff=cutoff):
        conds = [abs(h) < cutoff for h in hs_list[0]]
        for h, cond in zip(hs_list[0], conds): h[cond]=0
        for s, cond in zip(hs_list[1], conds): s[cond]=0

    @staticmethod
    def dense2csr(hs_list):
        hs_list[0] = [csr_matrix(mat) for mat in hs_list[0]]
        hs_list[1] = [csr_matrix(mat) for mat in hs_list[1]]

    def align_bf(self, bf):
        h1 = self.selfenergies[0].h_ii
        h_qii, s_qii = self.hs_list_ii
        h_qij, s_qij = self.hs_list_ij
        n = len(h_qii)
        diff = (h_qii[0][bf, bf] - h1[bf, bf].real) / s_qii[0][bf, bf]
        for q in range(n):
            h_qii[q] -= diff * s_qii[q]
        for q in range(n-1):
            h_qij[q] -= diff * s_qij[q]

    def get_g1N(self, energy):
        #
        if energy != self.energy:
            self.energy = energy

            mat_lists = self._get_mat_lists(energy)

            self.g1N = recursive_gf(*mat_lists)#_ii, mat_list_ij, mat_list_ji)

        return self.g1N


    def get_transmission(self, energies, T_e=None):

        # xp = xp.get_array_module(self.hs_list_ii[0][0])

        if T_e is None:
            T_e = np.zeros(len(energies))

        if self.parameters['gpu'] is False:
            # It is not convenient to spilt calculation of selfenergies
            # and recursive greenfuncition on same CPU.
            for e, energy in enumerate(energies):

                gr_1N = self.get_g1N(energy)
                ga_1N = xp.conj(gr_1N.T)
                lambda1_11 = self.selfenergies[0].get_lambda(energy)
                lambda2_NN = self.selfenergies[1].get_lambda(energy)
                T_e[e] = xp.trace(lambda1_11.dot(gr_1N).dot(lambda2_NN).dot(ga_1N)).real

        else:

            import threading
            # Threading calculator
            from .tcalc import Tcalc
            calc = Tcalc(self, energies, T_e)
            # Thread on CPU for computation of selfenergies
            bcs = threading.Thread(target=calc.bcs)
            # Thread on CPU for offloading recursive greenfuncition on GPU
            rec = threading.Thread(target=calc.rec)
            # bcs MUST start before rec (see locks)
            bcs.start()
            rec.start()
            # Wait for GPU.
            rec.join()

        return T_e

    def retarded(self, energy, trace=False, diag=False):

        gr_1N, G_qii, G_qij, G_qji = recursive_gf(*self._get_mat_lists(energy),
                                                   dos=True)

        return self._return(G_qii, G_qij, G_qji, trace=trace, diag=diag)


    def lesser(self, energy, fL=None, fR=None, trace=False, diag=False):

        s_in = [None for _ in range(self.N)]

        # Allow to consider separate channels
        if fL is not None:
            s_in[0] = self.selfenergies[0].get_lambda(energy) * fL
        if fR is not None:
            s_in[-1] = self.selfenergies[1].get_lambda(energy) * fR

        Gn_qii, Gn_qij, Gn_qji = recursive_gf(*self._get_mat_lists(energy),
                                               s_in=s_in, dos=True)

        return self._return(Gn_qii, Gn_qij, Gn_qji, trace=trace, diag=diag)



    def apply_retarded(self, energy, X_qii, X_qij, X_qji):#, X_1N=None):
        """Apply retarded Green function to X in tridiagonal form."""

        G_qii, G_qij, G_qji = self.retarded(energy)

        GX_qii = multiply(G_qii, G_qij, G_qji, X_qii, X_qij, X_qji)

        return GX_qii

    def apply_overlap(self, energy, trace=False, diag=False):
        S_qii = self.hs_list_ii[1]
        S_qij = self.hs_list_ij[1]
        S_qji = self.hs_list_ji[1]

        GS_qii = self.apply_retarded(energy, S_qii, S_qij, S_qji)#, S_1N)

        return self._return(GS_qii, trace=trace, diag=diag)


    def dos(self, energies, dos_e=None):
        if dos_e is None:
            dos_e = np.zeros_like(energies)
        for e, energy in enumerate(energies):
            GS = self.apply_overlap(energy, trace=True)
            dos_e[e] = - GS.imag / np.pi
        return dos_e


    def pdos(self, energies, bfs, sum=True):
        # Sum over basis functions bfs
        if sum:
            get = lambda a: - a.imag.take(bfs).sum() / np.pi
            pdos_e = np.zeros_like(energies)
        # Basis funtion's pdos
        else:
            get = lambda a: - a.imag.take(bfs) / np.pi
            pdos_e = np.zeros((len(energies),len(bfs)))
        for e, energy in enumerate(energies):
            GS_i = self.apply_overlap(energy, diag=True)
            pdos_e[e] = get(GS_i)
        return pdos_e

    def get_spectrals(self, energy):
        """Specral density of states incoming from leads."""
        A1_qii, A1_qij, A1_qji = self.lesser(energy, fL=1, fR=0)
        A2_qii, A2_qij, A2_qji = self.lesser(energy, fL=0, fR=1)
        A1_qmm = A1_qii, A1_qij, A1_qji
        A2_qmm = A2_qii, A2_qij, A2_qji

        return A1_qmm, A2_qmm

    def add_screening(self, V):
        if not hasattr(self, 'V'):
            self.V = np.zeros(self.nbf)
        assert V.size == self.nbf
        #Add screening and remove (if exists) current.
        h_qii = self.hs_list_ii[0]
        if sum(self.V) != 0:
            add_diagonal(h_qii, - self.V)
        self.V[:] = V
        add_diagonal(h_qii, self.V)


    def remove_screening(self):
        h_qii = self.hs_list_ii[0]
        add_diagonal(h_qii, -self.V)
        self.V[:] = 0.

    ## Helper functions

    @property
    def H(self):
        H_qii = self.hs_list_ii[0]
        H_qij = self.hs_list_ij[0]
        H_qji = self.hs_list_ji[0]
        return H_qii, H_qij, H_qji
    @H.setter
    def H(self,H):
        pass

    @property
    def S(self):
        S_qii = self.hs_list_ii[1]
        S_qij = self.hs_list_ij[1]
        S_qji = self.hs_list_ji[1]
        return S_qii, S_qij, S_qji
    @S.setter
    def S(self,S):
        pass

    def _get_mat_lists(self, energy):
        z = energy + self.eta * 1.j
        if self.selfenergies:
            sigma_L = self.selfenergies[0].retarded(energy)
            sigma_R = self.selfenergies[1].retarded(energy)
        else:
            sigma_L = None
            sigma_R = None
        # mat_list_ii, mat_list_ij, mat_list_ji

        if self.parameters['gpu'] is True:
            sigma_L = xp.asarray(sigma_L)
            sigma_R = xp.asarray(sigma_R)

        return get_mat_lists(z, self.hs_list_ii, self.hs_list_ij,
                             self.hs_list_ji, sigma_L, sigma_R)


    def _return(self, *args, trace, diag):

        if trace:
            return sum(A.trace() for A in args[0])
        elif diag:
            return get_diagonal(args[0])
        else:
            return args



from numba import vectorize

from qtpyt import xp

dagger = lambda mat: xp.conj(mat.T)

target = 'cpu' if xp.__name__=='numpy' else 'cuda'

kwargs = {'target':target}
if target == 'cpu': kwargs['fastmath'] = True

# @vectorize(['c16(c16,c16,c16)'],**kwargs)
def _mul(z,h,s):
    return z*s-h    

# @vectorize(['c16(c16,c16,c16)'],**kwargs)
def _mulc(z,h,s):
    return z*s.conj()-h.conj()
    # return z*s.conjugate()-h.conjugate()

def get_lambda(sigma):
    return 1.j * (sigma - sigma.T.conj())

def dotdiag(x, y, res):
    for i in range(x.shape[0]):
        res[i] += xp.sum(x[i,:] * y[:,i])

def dottrace(x, y, res=0.):
    for i in range(x.shape[0]):
        res += xp.sum(x[i,:] * y[:,i])
    return res



from setuptools import setup, Extension
from Cython.Build import cythonize


ext_modules = [
    Extension(
        "_recursive",
        ["_recursive.pyx"],
        extra_compile_args=["-O3", "-ffast-math", "-march=native", "-fopenmp"],
        extra_link_args=['-fopenmp'],
    )
]

setup(
    ext_modules = cythonize(ext_modules,
                            compiler_directives={"language_level": "3",
                                                 "boundscheck": False,
                                                 "wraparound": False})
 )



import numpy as np

#### NeighborList
from ase.data import covalent_radii
from ase.neighborlist import NeighborList
import ase.neighborlist


def get_neighbors(atoms):

    cov_radii = [covalent_radii[a.number] for a in atoms]
    nl = NeighborList(cov_radii, bothways = True, self_interaction = False)
    nl.update(atoms)

    nlists = []
    for a in atoms:
        nlists.append(nl.get_neighbors(a.index)[0])

    return nlists



import numpy as np
from functools import singledispatch
#
from .recursive import multiply, get_diagonal
from qtpyt.greenfunction import RecursiveGF, GreenFunction
from qtpyt.tools import dagger
# NeighborList
from qtpyt.tk_atoms import get_neighbors

def _orbital_current_gf(A_mm, H_mm, index_i, index_j,
                        A_dag_mm, H_dag_mm):
    ''' Reference bond current formula:
    Crossed graphene nanoribbons as beam splitters and mirrors
    for electron quantum optics.
    Sofia Sanz, et al.
    '''
    ni = len(index_i)
    nj = len(index_j)
    J_ij = np.zeros((ni,nj))
    for i,j in np.ndindex(ni,nj):
        ii = index_i[i]
        jj = index_j[j]
        J_ij[i,j] = np.imag(H_dag_mm[jj,ii]*A_mm[ii,jj] -
                            H_mm[ii,jj]*A_dag_mm[jj,ii])
    return J_ij

def _orbital_current_rgf(A_xqmm, H_xqmm, index_i, index_j):
    A_qii, A_qij, A_qji = A_xqmm
    H_qii, H_qij, H_qji = H_xqmm
    sizes = np.cumsum([H_mm.shape[0] for H_mm in H_qii])
    # Indices of atoms in recursive partition
    # np.searchsorted(..,side='right') because
    # if indices_i[0] == sizes[q=0], then atom i
    # goes to next partition q=1
    #
    qi = np.searchsorted(sizes, index_i, side='right')
    qj = np.searchsorted(sizes, index_j, side='right')

    # Check
    assert len(set(qi))==1
    assert len(set(qj))==1
    # If OK, all orbitals belong to same partition
    qi = qi[0]
    qj = qj[0]

    M_q = np.insert(sizes,0,0)[:-1]
    index_i = np.asarray(index_i) - M_q[qi]
    index_j = np.asarray(index_j) - M_q[qj]

    if qi==qj:
        A_mm = A_qii[qi]
        H_mm = H_qii[qi]
        A_dag_mm = A_mm
        H_dag_mm = H_mm
    elif qj==qi+1:
        A_mm = A_qij[qi]
        H_mm = H_qij[qi]
        A_dag_mm = A_qji[qi]
        H_dag_mm = H_qji[qi]
    elif qi==qj+1:
        A_mm = A_qji[qj]
        H_mm = H_qji[qj]
        A_dag_mm = A_qij[qj]
        H_dag_mm = H_qij[qj]

    return _orbital_current_gf(A_mm, H_mm, index_i, index_j,
                               A_dag_mm, H_dag_mm)

def orbital_current(A, H, index_i, index_j):
    if isinstance(A, (list,tuple)):
        return _orbital_current_rgf(A, H, index_i, index_j)
    else:
        return _orbital_current_gf(A, H, index_i, index_j, A, H)

def bond_current(A, H, bfs_ai, nlists):
    J_aa = [[None for _ in nlist] for nlist in nlists]
    for a0, nlist in enumerate(nlists):
        index_i = bfs_ai[a0]
        for j, a1 in enumerate(nlist):
            index_j = bfs_ai[a1]
            J_aa[a0][j] = orbital_current(A, H,
                                          index_i,
                                          index_j).sum()
    return J_aa

def quiver_plot(atoms, J_aa, nlists, **kwargs):
    ''' Get arrows info
    X - origins
    U - vectors with legth of corresponding bond
    W - weigths
    '''
    X = []
    U = []
    W = []
    for a in atoms:
        nlist = nlists[a.index]
        pos = np.tile(a.position, (len(nlist), 1))
        dist = (atoms[nlist].positions - a.position) #\
                #/ atoms.get_distances(a.index, nlist)[:,None]
        X.extend(pos.tolist())
        U.extend(dist.tolist())
        W.extend(J_aa[a.index])

    X = np.asarray(X)
    U = np.asarray(U)
    W = np.asarray(W)
    Wneg = W<0
    # Shift arrow to moddle of plot
    X[Wneg] += U[Wneg]/2
    # Reverse arrow
    U[Wneg] *= -1
    W = np.abs(W)
    return X, U, W

def normalize_weigths(W, normalize=False, vmin=None, vmax=None, clip=False):
    if normalize:
        W = (W-W.min())/(W.max()-W.min())
    if clip:
        W = np.clip(W, vmin, vmax)
    return W


class DensityCurrent(object):

    def __init__(self, GF, atoms, bfs_ai):
        self.GF = GF
        self.atoms = atoms
        self.bfs_ai = bfs_ai
        self.energy = None
        self._set_nlists()
        self.indices = None

    def __call__(self, energy):
        # Spectral functions
        if energy != self.energy:
            GF = self.GF
            self.A1, self.A2 = GF.get_spectrals(energy)
        bfs_ai = self._get_lists('bfs_ai')
        nlists = self._get_lists('nlists')
        J1_aa = bond_current(self.A1, GF.H, bfs_ai, nlists)
        J2_aa = bond_current(self.A2, GF.H, bfs_ai, nlists)

        return J1_aa, J2_aa, nlists

    def _get_lists(self, name):
        attr = getattr(self, name)
        if self.indices is None:
            return attr
        else:
            return [l for a,l in enumerate(attr) if a in self.indices]

    def _set_nlists(self):
        atoms = self.atoms
        pbc = atoms.pbc.copy()
        # Hack to avoid arrow between boundaries
        atoms.set_pbc(False)
        self.nlists = get_neighbors(atoms)
        atoms.set_pbc(pbc)

    def add_neighbors(self, index, neighbors):
        if isinstance(neighbors, int):
            neighbors = [neighbors]
        elif isinstance(neighbors, np.ndarry):
            neighbors = neighbors.tolist()
        self.nlists[index].extend(neighbors)
        for neighbor in neighbors:
            self.nlists[neighbor].append(index)

    def pop_neighbors(self, index, neighbors):
        if isinstance(neighbors, int):
            neighbors = [neighbors]
        elif isinstance(neighbors, np.ndarry):
            neighbors = neighbors.tolist()
        for neighbor in neighbors:
            ii = self.nlists[index].index(neighbor)
            jj = self.nlists[neighbor].index(index)
            self.nlists[index].pop(ii)
            self.nlists[neighbor].pop(jj)

    def display(self, J_aa, nlists, projection='2d', plane='xy', scale=2,
                normalize=True, vmin=None, vmax=None, clip=False):

        from mpl_toolkits.mplot3d import Axes3D
        from matplotlib import pyplot as plt
        import matplotlib.cm as cm
        # from matplotlib.colors import Normalize
        atoms = self.atoms

        X, U, W = quiver_plot(atoms, J_aa, nlists)
        W = normalize_weigths(W, normalize, vmin, vmax, clip)
        colormap = cm.YlOrBr

        fig = plt.figure(figsize=(atoms.cell[0,0]/scale,
                                  atoms.cell[1,1]/scale))
        if projection is '3d':
            ax = fig.gca(projection=projection)
            ax.quiver(X[:,0],X[:,1],X[:,2],
                      U[:,0],U[:,1],U[:,2],
                      color=colormap(W))
        else:
            ax = fig.gca()
            a = 'xyz'.index(plane[0])
            b = 'xyz'.index(plane[1])
            ax.quiver(X[:,a],X[:,b],
                      U[:,a],U[:,b],
                      color=colormap(W))
            ax.scatter(atoms.positions[:,a], atoms.positions[:,b])
        return ax



import numpy as np
from scipy.spatial.distance import cdist

from .kpts import (
                   monkhorst_pack, apply_inversion_symm, 
                   build_partial_kpts, partial_bloch_sum)

def remove_pbc(basis, A_kMM, direction='x', eps=-1e-3):
    """Mask matrix elements above diagonal connecting neighbors cells
    in parallel direction.

    """

    # Transport direction
    p_dir = 'xyz'.index(direction)

    L = basis.atoms.cell[p_dir, p_dir]

    cutoff = L - eps
    # Coordinates of central unit cell i (along transport)
    centers_p_i = basis.centers[:, p_dir]
    # Coordinates of neighbooring unit cell j
    centers_p_j = centers_p_i + L
    # Distance between j atoms and i atoms
    dist_p_ji = np.abs(centers_p_j[:, None] - centers_p_i[None, :])
    # Mask j atoms farther than L
    mask_ji = (dist_p_ji > cutoff).astype(A_kMM.dtype)

    A_kMM *= mask_ji[None, :]


def oder_indices(basis, N, order='xyz', positions=None):
    repeated = basis.repeat(N)
    if positions is not None:
        idx = cdist(repeated.centers, positions).argmin(1) 
        return repeated.get_index(idx)
    else:
        idx = repeated.argsort(order)
        return repeated.get_indices(idx)


def prepare_leads_matrices(basis, H_kMM, S_kMM, size, offset=(0.,0.,0.), direction='x'):
    """Prepare input matrices for PrincipalLayer.
    
    Args:
        basis : basis function descriptor.
        H_kMM : Hamiltonian matrices in k-space.
        S_kMM : Overlap matrices in k-space.
        size, offset : Monkhorst-Pack used to sample Brillouin Zone.
    """

    kpts = monkhorst_pack(size, offset)
    if kpts.shape[0] > H_kMM.shape[0]:
        # Switch to irreducible k-points
        kpts = apply_inversion_symm(kpts)[0]

    kpts_p, kpts_t = build_partial_kpts(size, offset, direction)

    p_dir = 'xyz'.index(direction)

    R = [0,0,0]
    H_kii = partial_bloch_sum(H_kMM, kpts, R, kpts_p, kpts_t)
    S_kii = partial_bloch_sum(S_kMM, kpts, R, kpts_p, kpts_t)

    R[p_dir] = 1
    H_kij = partial_bloch_sum(H_kMM, kpts, R, kpts_p, kpts_t)
    S_kij = partial_bloch_sum(S_kMM, kpts, R, kpts_p, kpts_t)

    remove_pbc(basis, H_kij, direction)
    remove_pbc(basis, S_kij, direction)

    return kpts_t, H_kii, S_kii, H_kij, S_kij


import cupy as cp
from numba import cuda
from math import ceil

threadsperblock = (16,16)

def get_blockspergrid(a, threadsperblock):
    blockspergrid_x = ceil(a.shape[0] / threadsperblock[0])
    blockspergrid_y = ceil(a.shape[1] / threadsperblock[1])
    blockspergrid = (blockspergrid_x, blockspergrid_y)
    return blockspergrid


@cuda.jit
def _geamT(z, h, s, m):
    x, y = cuda.grid(2)
    if x < h.shape[0] and y < h.shape[1]:
        m[y, x] = z * s[x, y] - h[x, y]


@cuda.jit
def _geamC(z, h, s, m):
    x, y = cuda.grid(2)
    if x < h.shape[0] and y < h.shape[1]:
        m[x, y] = z * s[x, y].conjugate() - h[x, y].conjugate()


@cuda.jit
def _get_lambda(sigma, lambd):
    x, y = cuda.grid(2)
    if x < sigma.shape[0] and y < sigma.shape[1]:
        lambd[y, x] = 1.j * (sigma[x, y] - sigma[y, x].conjugate())


def geamT(z, h, s, m=None):
    assert h.flags.c_contiguous & s.flags.c_contiguous
    if m is not None:
        assert m.flags.f_contiguous
    else:
        m = cp.empty(h.shape, complex, order='F')
    blockspergrid = get_blockspergrid(h, threadsperblock)
    _geamT[blockspergrid, threadsperblock](z, h, s, m)
    return m


def geamC(z, h, s, m=None):
    assert h.flags.c_contiguous & s.flags.c_contiguous
    if m is not None:
        assert m.flags.f_contiguous
    else:
        m = cp.empty(h.shape, complex, order='F')
    blockspergrid = get_blockspergrid(h, threadsperblock)
    _geamC[blockspergrid, threadsperblock](z, h, s, m)
    return m


def get_lambda(sigma, lambd=None):
    assert sigma.flags.c_contiguous
    if lambd is not None:
        assert lambd.flags.f_contiguous
    else:
        lambd = cp.empty(sigma.shape, complex, order='F')
    blockspergrid = get_blockspergrid(sigma, threadsperblock)
    _get_lambda[blockspergrid, threadsperblock](sigma, lambd)
    return lambd


if __name__ == '__main__':
    z = complex(2.)

    h = cp.random.random((10,10)).astype(complex)
    s = cp.random.random((10,10)).astype(complex)
    m = cp.empty((10,10), complex, order='F')

    blockspergrid = get_blockspergrid(h, threadsperblock)

    geamT(z, h, s, m)
    expected = z*s.T-h.T
    assert cp.allclose(expected,m)

    geamC(z, h, s, m)
    expected = z*s.conj()-h.conj()
    assert cp.allclose(expected,m)

    sigma = cp.random.random((10,10)).astype(complex)
    lambd = cp.empty((10,10), complex, order='F')

    get_lambda(sigma, lambd)
    expected = 1.j * (sigma - sigma.T.conj()).T
    assert cp.allclose(expected,lambd)


import numpy as np

# Taken from gpaw.lcao.tools.py
def tri2full(H_nn, UL='L', map=np.conj):
    """Fill in values of hermitian or symmetric matrix.

    Fill values in lower or upper triangle of H_nn based on the opposite
    triangle, such that the resulting matrix is symmetric/hermitian.

    UL='U' will copy (conjugated) values from upper triangle into the
    lower triangle.

    UL='L' will copy (conjugated) values from lower triangle into the
    upper triangle.

    The map parameter can be used to specify a different operation than
    conjugation, which should work on 1D arrays.  Example::

      def antihermitian(src, dst):
            np.conj(-src, dst)

      tri2full(H_nn, map=antihermitian)

    """
    N, tmp = H_nn.shape
    assert N == tmp, 'Matrix must be square'
    if UL != 'L':
        H_nn = H_nn.T

    for n in range(N - 1):
        map(H_nn[n + 1:, n], H_nn[n, n + 1:])




import numpy as np
import ctypes

mkl = ctypes.cdll.LoadLibrary('libmkl_rt.so')

i4 = ctypes.c_int
f8 = ctypes.c_double
c16 = ctypes.c_longdouble


def dot(a, b, c, alpha=1., beta=0.):

    Order = i4(101)
    TransA = i4(111)
    TransB = i4(111)

    a_ptr = np.ctypeslib.as_ctypes(a)
    b_ptr = np.ctypeslib.as_ctypes(b)
    c_ptr = np.ctypeslib.as_ctypes(c)

    m = i4(a.shape[0])
    n = i4(b.shape[1])
    k = i4(a.shape[1])

    lda = k
    ldb = n
    ldc = n

    mkl.cblas_dgemm(Order,TransA,TransB,m,n,k,f8(alpha),a_ptr,lda,b_ptr,ldb,f8(beta),c_ptr,ldc)

    return c


def solve(a, b, overwrite_b=False):
    """Solve the equation:   
        Ax=B
    
    Returns: A^(-1)B
    """
    Order = i4(102) if a.flags.f_contiguous else i4(101)

    _b = b if overwrite_b else b.copy(order='A')

    a_ptr=a.ctypes.data_as(ctypes.POINTER(c16))
    b_ptr=_b.ctypes.data_as(ctypes.POINTER(c16))

    n = i4(a.shape[0])
    nrhs = i4(b.shape[1])
    ipiv = np.empty(a.shape[0],np.int32)
    ipiv_ptr = ipiv.ctypes.data_as(ctypes.POINTER(i4))

    lda = n
    ldb = n if b.flags.f_contiguous else nrhs

    mkl.lapacke_zgesv(Order,n,nrhs,a_ptr,lda,ipiv_ptr,b_ptr,ldb)

    return _b


def lrsolve(a, b):
    """Solve the equation:   
        Ax=B 
        A^(H)y=x

    Returns: A^(-1)BA^(H)^(-1)
    """
    Order = i4(102) if a.flags.f_contiguous else i4(101)
    
    a_ptr = a.ctypes.data_as(ctypes.POINTER(c16))
    b_ptr = b.ctypes.data_as(ctypes.POINTER(c16))

    m = i4(a.shape[0])
    n = i4(a.shape[1])
    lda = m if a.flags.f_contiguous else n

    ipiv = np.empty(a.shape[0], np.int32)
    ipiv_ptr = ipiv.ctypes.data_as(ctypes.POINTER(c16))

    mkl.lapacke_zgetrf(Order,m,n,a_ptr,lda,ipiv_ptr)

    trans = ctypes.c_char(b'N')
    nrhs = i4(b.shape[1])
    ldb = m if b.flags.f_contiguous else nrhs

    mkl.lapacke_zgetrs(Order,trans,n,nrhs,a_ptr,lda,ipiv_ptr,b_ptr,ldb)

    np.conj(a, out=a)

    c = b.copy(order='C') if a.flags.f_contiguous else b.copy(order='F')
    c_ptr = c.ctypes.data_as(ctypes.POINTER(c16))

    mkl.lapacke_zgetrs(Order,trans,n,nrhs,a_ptr,lda,ipiv_ptr,c_ptr,ldb)

    return c

def tsolve(a, b, c):
    """Solve the equation:   
        Ax=B 
        A^(H)y=C

    Returns: A^(-1)B, A^(H)^(-1)C
    """
    Order = i4(102) if a.flags.f_contiguous else i4(101)

    a_ptr = a.ctypes.data_as(ctypes.POINTER(c16))
    b_ptr = b.ctypes.data_as(ctypes.POINTER(c16))
    c_ptr = c.ctypes.data_as(ctypes.POINTER(c16))

    m = i4(a.shape[0])
    n = i4(a.shape[1])
    lda = m if a.flags.f_contiguous else n

    ipiv = np.empty(a.shape[0], np.int32)
    ipiv_ptr = ipiv.ctypes.data_as(ctypes.POINTER(c16))

    mkl.lapacke_zgetrf(Order,m,n,a_ptr,lda,ipiv_ptr)

    trans = ctypes.c_char(b'N')
    nrhs = i4(b.shape[1])
    ldb = m if b.flags.f_contiguous else nrhs

    mkl.lapacke_zgetrs(Order,trans,n,nrhs,a_ptr,lda,ipiv_ptr,b_ptr,ldb)

    trans = ctypes.c_char(b'C')
    nrhs = i4(c.shape[1])
    ldc = m if c.flags.f_contiguous else nrhs

    mkl.lapacke_zgetrs(Order,trans,n,nrhs,a_ptr,lda,ipiv_ptr,c_ptr,ldc)

    return b, c


import numpy as np

from qtpyt.kernels.linalg_cpu import (solve, lrsolve, tsolve, dot)

def test_dot():
    
    a = np.ones((50,100))
    b = np.ones((100,200))
    c = np.empty((50,200))

    expected = np.dot(a,b)

    c = dot(a,b,c,beta=0.)

    np.testing.assert_allclose(c, expected)


def test_solve():
    
    a = np.random.random((100,100)) + 1.j*np.random.random((100,100))
    b = np.random.random((100,200)) + 1.j*np.random.random((100,200))

    expected = np.linalg.solve(a,b)

    a_f = a.copy(order='F')
    b_f = b.copy(order='F')

    c = solve(a,b)

    np.testing.assert_allclose(c, expected)

    c = solve(a_f,b_f)

    np.testing.assert_allclose(c, expected)


def test_lrsolve():

    a = np.random.random((100,100)) + 1.j*np.random.random((100,100))
    b = np.random.random((100,100)) + 1.j*np.random.random((100,100))

    tmp = np.linalg.solve(a,b)
    expected = np.linalg.solve(a.conj(),tmp.T).T

    a_f = a.copy(order='F')
    b_f = b.copy(order='F')

    c = lrsolve(a,b)

    np.testing.assert_allclose(c, expected)

    c = lrsolve(a_f,b_f)

    np.testing.assert_allclose(c, expected)


def test_tsolve():

    a = np.random.random((100,100)) + 1.j*np.random.random((100,100))
    b = np.random.random((100,100)) + 1.j*np.random.random((100,100))
    c = np.random.random((100,100)) + 1.j*np.random.random((100,100))

    expected1 = np.linalg.solve(a,b)
    expected2 = np.linalg.solve(a.T.conj(),c)

    a_f = a.copy(order='F')
    b_f = b.copy(order='F')
    c_f = c.copy(order='F')

    d1, d2 = tsolve(a,b,c)

    np.testing.assert_allclose(d1, expected1)
    np.testing.assert_allclose(d2, expected2)

    d1, d2 = tsolve(a_f,b_f,c_f)

    np.testing.assert_allclose(d1, expected1)
    np.testing.assert_allclose(d2, expected2)


def time_dgemm():
    import timeit

    m = 1000
    k = 2000
    n = 700
    a = np.ones((m,k))
    b = np.ones((k,n))
    c = np.empty((m,n))

    elapsed = timeit.timeit("np.dot(a,b,out=c)", 
        setup='import numpy as np', globals=locals(), number=500)

    print("numpy dot :", elapsed)

    elapsed = timeit.timeit("dot(a,b,c,beta=0.)", 
        setup='from __main__ import dot', globals=locals(), number=500)
    
    print("This :", elapsed)


def time_solve():
    import timeit
    
    n = 1000
    a = np.random.random((n,n)).astype(complex)
    nrhs = 2000
    b = np.random.random((n,nrhs)).astype(complex)

    # elapsed = timeit.timeit("la.solve(a,b,overwrite_b=True,check_finite=False)", 
    #     setup='from scipy import linalg as la', globals=locals(), number=500)

    # print("scipy solve :", elapsed)

    elapsed = timeit.timeit("la.solve(a,b)", 
        setup='from numpy import linalg as la', globals=locals(), number=100)

    print("numpy solve :", elapsed)

    elapsed = timeit.timeit("solve(a,b,overwrite_b=True)", 
        setup='from __main__ import solve', globals=locals(), number=100)
    
    print("This :", elapsed)


def time_lrsolve():
    import timeit
    
    n = 1000
    a = np.random.random((n,n)).astype(complex)
    nrhs = 1000
    b = np.random.random((n,nrhs)).astype(complex)

    a_f = a.copy(order='F')
    b_f = b.copy(order='F')

    elapsed = timeit.timeit("tmp=la.solve(a,b);la.solve(a,tmp.T).T",
        setup='from numpy import linalg as la', globals=locals(), number=100)

    print("numpy solve :", elapsed)

    # elapsed = timeit.timeit("ainv=la.inv(a);ainv.dot(b).dot(ainv.T)",
    #     setup='from numpy import linalg as la', globals=locals(), number=100)

    # print("numpy inv :", elapsed)

    elapsed = timeit.timeit("lrsolve(a_f,b_f)",
        setup='from __main__ import lrsolve', globals=locals(), number=100)
    
    print("This solve:", elapsed)


def time_tsolve():
    import timeit
    
    n = 1000
    a = np.random.random((n,n)).astype(complex)
    nrhs = 1000
    b = np.random.random((n,nrhs)).astype(complex)
    c = np.random.random((n,nrhs)).astype(complex)

    a_f = a.copy(order='F')
    b_f = b.copy(order='F')
    c_f = c.copy(order='F')

    elapsed = timeit.timeit("expected1=la.solve(a,b);expected2=la.solve(a.T.conj(),c)",
        setup='from numpy import linalg as la', globals=locals(), number=100)

    print("numpy solve :", elapsed)

    elapsed = timeit.timeit("d1,d2=tsolve(a_f,b_f,c_f)",
        setup='from __main__ import tsolve', globals=locals(), number=100)
    
    print("This solve:", elapsed)


if __name__ == '__main__':
    # time_dgemm()
    time_solve()
    time_lrsolve()
    time_tsolve()


import numpy as np
import numba
from numba import prange

@numba.njit(parallel=True,fastmath=True)
def _geamT(z, h, s, m):
    for i in prange(h.shape[0]):
        for j in range(h.shape[1]):
            m[j,i] = z * s[i,j] - h[i,j] 


@numba.njit(parallel=True,fastmath=True)
def _geamC(z, h, s, m):
    for i in prange(h.shape[0]):
        for j in range(h.shape[1]):
            m[i,j] = z * s[i,j].conjugate() - h[i,j].conjugate() 


@numba.njit(parallel=False,fastmath=True)
def _get_lambda(sigma, lambd):
    for i in range(sigma.shape[0]):
        for j in range(sigma.shape[1]):
            lambd[j,i] = 1.j * (sigma[i,j] - sigma[j,i].conjugate())


def geamT(z, h, s, m=None):
    assert h.flags.c_contiguous & s.flags.c_contiguous
    if m is not None:
        assert m.flags.f_contiguous
    else:
        m = np.empty(h.shape, complex, order='F')
    _geamT(z, h, s, m)
    return m


def geamC(z, h, s, m=None):
    assert h.flags.c_contiguous & s.flags.c_contiguous
    if m is not None:
        assert m.flags.f_contiguous
    else:
        m = np.empty(h.shape, complex, order='F')
    _geamC(z, h, s, m)
    return m


def get_lambda(sigma, lambd=None):
    assert sigma.flags.c_contiguous
    if lambd is not None:
        assert lambd.flags.f_contiguous
    else:
        lambd = np.empty(sigma.shape, complex, order='F')
    _get_lambda(sigma, lambd)
    return lambd



if __name__ == '__main__':
    z = complex(2.)

    h = np.random.random((900,900)).astype(complex)
    s = np.random.random((900,900)).astype(complex)
    m = np.empty((900,900), complex, order='F')

    geamT(z, h, s, m)
    expected = z*s.T-h.T
    assert np.allclose(expected,m)

    geamC(z, h, s, m)
    expected = z*s.conj()-h.conj()
    assert np.allclose(expected,m)

    sigma = np.random.random((900,900)).astype(complex)
    lambd = np.empty((900,900), complex, order='F')

    get_lambda(sigma, lambd)
    expected = 1.j * (sigma - sigma.T.conj()).T
    assert np.allclose(expected,lambd)



import cupy as cp

from qtpyt.kernels.linalg_gpu import (solve, lrsolve, tsolve)


def test_solve():
    a = cp.random.random((100,100)) + 1.j*cp.random.random((100,100))
    b = cp.random.random((100,100)) + 1.j*cp.random.random((100,100))

    a_f = cp.asfortranarray(a)
    b_f = cp.asfortranarray(b)

    c = solve(a_f,b_f)

    expected = cp.linalg.solve(a,b)

    assert cp.allclose(c, expected)


def test_lrsolve():
    a = cp.random.random((100,100)) + 1.j*cp.random.random((100,100))
    b = cp.random.random((100,100)) + 1.j*cp.random.random((100,100))

    a_f = cp.asfortranarray(a)
    b_f = cp.asfortranarray(b)

    c = lrsolve(a_f,b_f)

    tmp = cp.linalg.solve(a,b)
    expected = cp.linalg.solve(a.conj(),tmp.T).T

    assert cp.allclose(c, expected)


def test_tsolve():
    a = cp.random.random((100,100)) + 1.j*cp.random.random((100,100))
    b = cp.random.random((100,100)) + 1.j*cp.random.random((100,100))
    c = cp.random.random((100,100)) + 1.j*cp.random.random((100,100))

    a_f = cp.asfortranarray(a)
    b_f = cp.asfortranarray(b)
    c_f = cp.asfortranarray(c)

    d_f1, d_f2 = tsolve(a_f,b_f,c_f)

    expected1 = cp.linalg.solve(a,b)
    expected2 = cp.linalg.solve(a.T.conj(),c)

    assert cp.allclose(d_f1, expected1)
    assert cp.allclose(d_f2, expected2)   


import numpy as np
from scipy import linalg as la

#Cutoff for cutting block matrices
from .tridiagonal import cutoff
from qtpyt.tools import dagger

# import cupy as cp
from transport import xp

def recursive_gf(mat_list_ii, mat_list_ij, mat_list_ji, s_in=None, dos=False):

    # xp = cp.get_array_module(mat_list_ii[0][0])

    N = len(mat_list_ii)
#     mat_shapes = [mat.shape[0] for mat in mat_list_ii]

    # np.matrix alias
    m_qii = mat_list_ii
    m_qij = mat_list_ij
    m_qji = mat_list_ji

    # Left connected green's function
    grL_qii = [None for _ in range(N)]
    # Initalize
    grL_qii[0] = xp.linalg.inv(m_qii[0]) # ([eS-H]_11-Sigma_L)^-1
    # First row green's function
    gr_1i = grL_qii[0].copy()

    # assert xp == cp.get_array_module(gr_1i)

    # Left connected recursion
    for q in range(1, N):
        # Left
        grL_qii[q] = xp.linalg.inv(m_qii[q] - xp.dot(xp.dot(m_qji[q - 1],grL_qii[q - 1]),m_qij[q - 1]))
        # 1st row
        gr_1i = xp.dot(xp.dot(gr_1i,m_qij[q - 1]), grL_qii[q])

    # Only transport
    if not dos:
        # Return g1N
        return gr_1i

    # Full green's function
    Gr_qii = [None for _ in range(N)]
    Gr_qji = [None for _ in range(N-1)]
    Gr_qij = [None for _ in range(N-1)]
    # Initialize
    Gr_qii[-1] = grL_qii[-1] # G_NN = gL_NN = ([eS-H]_NN - [eS-H]_NN-1 * grL_N-1N-1 * [eS-H]_N-1N - Sigma_R)^-1

    # Full recursion
    for q in range(N-2, -1, -1):
        Gr_qji[q] = - xp.dot(Gr_qii[q + 1],xp.dot(m_qji[q],grL_qii[q]))
        Gr_qij[q] = - xp.dot(grL_qii[q],xp.dot(m_qij[q],Gr_qii[q + 1]))
        Gr_qii[q] = grL_qii[q] - xp.dot(grL_qii[q],xp.dot(m_qij[q],Gr_qji[q]))

    # Return retarded
    if s_in is None:
        # DOS
        return gr_1i, Gr_qii, Gr_qij, Gr_qji

    # Electron correlation function
    if isinstance(s_in, list):

        gnL_qii = [None for _ in range(N)]
        # Initalize
        gnL_qii[0] = grL_qii[0] @ s_in[0] @ np.conj(grL_qii[0])

        # Left connected recursion
        for q in range(1, N):
            sl = m_qji[q - 1] @ gnL_qii[q - 1] @ np.conj(m_qij[q - 1])
            if s_in[q] is not None: sl += s_in[q]
            gnL_qii[q] = grL_qii[q] @ sl @ np.conj(grL_qii[q])

        Gn_qii = [None for _ in range(N)]
        Gn_qij = [None for _ in range(N-1)]
        Gn_qji = [None for _ in range(N-1)]
        # Initialize
        Gn_qii[-1] = gnL_qii[-1]

        # Full recursion
        for q in range(N-2, -1, -1):
            a = Gn_qii[q + 1] @ np.conj(m_qji[q]) @ np.conj(grL_qii[q])
            Gn_qji[q] = - Gr_qii[q + 1] @ m_qji[q] @ gnL_qii[q] - a
            Gn_qii[q] = gnL_qii[q] + \
                        grL_qii[q] @ m_qij[q] @  a - \
                        gnL_qii[q] @ np.conj(m_qij[q]) @ np.conj(Gr_qji[q]) - \
                        Gr_qij[q] @ m_qji[q] @ gnL_qii[q]
            Gn_qij[q] = dagger(Gn_qji[q])

    # Return electron correlation
    return Gn_qii, Gn_qij, Gn_qji

    # # Right connected green's function
    # hgh_qii = [None for _ in range(N-1)]
    # # Initalize
    # grR_inv = m_qii[-1] # ([eS-H]_NN-Sigma_R)^-1
    #
    # # Right connected recursion
    # for q in range(N-2, -1, -1):
    #     # Left
    #     hgh_qii[q] = m_qij[q] @ la.solve(grR_inv, m_qji[q])
    #     grR_inv = m_qii[q] - hgh_qii[q]
    #
    # # Full green's function
    # Gr_qii = [None for _ in range(N)]
    # Gr_qii[-1] = grL_qii[-1]       #Actual gNN
    # Gr_qii[0]  = la.inv(grR_inv)   #Actual g11
    # for q in range(1, N-1):
    #     Gr_qii[q] = grL_qii[q] @ la.inv(
    #     np.eye(mat_shapes[q]) - grL_qii[q] @ hgh_qii[q])
    #
    # return Gr_qii, None, None
    #

def get_mat_lists(z, hs_list_ii, hs_list_ij, hs_list_ji, sigma_L=None, sigma_R=None):

    mat_list_ii = []
    mat_list_ij = []
    mat_list_ji = []
    # Upper
    h_list_ij, s_list_ij = hs_list_ij
    for h_ij, s_ij in zip(h_list_ij,
                          s_list_ij):

        mat_list_ij.append(z * s_ij - h_ij)
    # Lower
    h_list_ji, s_list_ji = hs_list_ji
    for h_ji, s_ji in zip(h_list_ji,
                          s_list_ji):

        mat_list_ji.append(z * s_ji - h_ji)
    # Diagonal
    h_list_ii, s_list_ii = hs_list_ii
    for h_ii, s_ii in zip(h_list_ii,
                          s_list_ii):
        mat_list_ii.append(z * s_ii - h_ii)
    # Add selfenergies
    if sigma_L is not None:
        # Left
        mat_list_ii[0]  -= sigma_L
    if sigma_R is not None:
        # Right
        mat_list_ii[-1] -= sigma_R

    return mat_list_ii, mat_list_ij, mat_list_ji

def multiply(A_qii, A_qij, A_qji, B_qii, B_qij, B_qji):
    """Helper function to multiply two block tridiagonal
    matrices."""
    N = len(A_qii)
    # Diagonal sum
    AB_qii = [xp.dot(a,b) for a,b in zip(A_qii,B_qii)]
    # Upper diagonal sum
    for q in range(N-1):
        AB_qii[q][:] += xp.dot(A_qij[q],B_qji[q])
    # Lower diagonal sum
    for q in range(1,N):
        AB_qii[q][:] += xp.dot(A_qji[q-1],B_qij[q-1])

    return AB_qii

def add_diagonal(A_qii, V):
    i0 = 0
    for A_ii in A_qii:
        i1 = i0 + len(A_ii)
        A_ii.flat[:: len(A_ii) + 1] += V[i0:i1]
        i0 = i1

def get_diagonal(A_qii):
    if xp.__name__ == 'cupy':
        get = lambda a: xp.diagonal(a).get()
    else:
        get = lambda a: xp.diagonal(a)
    nao = sum(len(A) for A in A_qii)
    A_i = np.zeros(nao, A_qii[0].dtype)
    # Loop over diagonal elements
    i0 = 0
    for A_ii in A_qii:
        i1 = i0 + len(A_ii)
        A_i[i0:i1] = get(A_ii) #np.diagonal(A_ii) #.diagonal()
        i0 = i1
    return A_i



from qtpyt import xp

from qtpyt.abc._kernels import (_mul,_mulc)
from qtpyt.abc.greenfunction import GreenFunction


class SelfEnergy(GreenFunction):

    ids = ['left','right']
    
    def __init__(self, hs_ii, hs_im, selfenergies=[], eta=1e-5, id='left'):
        super().__init__(hs_ii[0], hs_ii[1], selfenergies, eta)

        assert id in SelfEnergy.ids, f"Invalid id for Principal self-energy. Choose between {PrincipalSelfEnergy.ids}"
        self.id = id
        if id == 'right':
            h_im = hs_im[0].swapaxes(0,1).conj()
            s_im = hs_im[1].swapaxes(0,1).conj()
            hs_im = (h_im, s_im)
        
        self.h_im = xp.asarray(hs_im[0]) 
        self.s_im = xp.asarray(hs_im[1])
        
        self.eta = eta
        self.bias = 0.
        self.energy = None

        self.sigma_mm = xp.empty(2*(self.nbf_m,), complex)

    @property
    def nbf_i(self):
        return self.h_im.shape[0]

    @property
    def nbf_m(self):
        return self.h_im.shape[1]

    def retarded(self, energy):
        """Return self-energy (sigma) evaluated at specified energy."""
        if energy != self.energy:
            # self.energy = energy # Done by Green's function.
            z = energy - self.bias + self.eta * 1.j
            tau_im = _mul(z,self.h_im,self.s_im)
            Ginv = self.get_Ginv(energy)
            a_im = xp.linalg.solve(Ginv, tau_im)
            tau_mi = _mulc(z,self.h_im.T,self.s_im.T)
            xp.dot(tau_mi, a_im, out=self.sigma_mm)

        return self.sigma_mm

    def get_lambda(self, energy):
        """Return the lambda (aka Gamma) defined by i(S-S^d).

        Here S is the retarded selfenergy, and d denotes the hermitian
        conjugate.
        """
        sigma_mm = self.retarded(energy)
        return 1.j * (sigma_mm - sigma_mm.T.conj())



from qtpyt import xp

if xp.__name__ == 'numpy':
    from .linalg_cpu import (solve, lrsolve, tsolve)
    from .hs_cpu import (geamC, geamT, get_lambda)

else:
    from .linalg_gpu import (solve, lrsolve, tsolve)
    from .hs_gpu import (geamC, geamT, get_lambda)  



import numpy as np
from functools import lru_cache
from .coupledhamiltonian import CoupledHamiltonian
from .block_tridiag import tridiagonalize, cutoff
from .recursive_gf import recursive_gf

class GreenFunction(CoupledHamiltonian):
    """Equilibrium retarded Green function."""

    def __init__(self, H, S=None, selfenergies=[], eta=1e-4):
        self.H = H
        self.S = S
        self.selfenergies = selfenergies
        self.eta = eta
        self.energy = None
        self.Ginv = np.empty(H.shape, complex)
        self.is_recursive = False

    def set_tridiag(self, calc, pl1, pl2=None, cutoff=cutoff):

        self.calc = calc
        self.hs_list_ii, self.hs_list_ij = tridiagonalize(
                                           calc, self.H, self.S,
                                           pl1, pl2, cutoff)

        for selfenergy in self.selfenergies:
            selfenergy.set_hs_im(selfenergy.nbf_i, selfenergy.nbf_i)

        self.g_1N = None
        self.is_recursive = True

    # @lru_cache(maxsize=None)
    def retarded(self, energy, inverse=False):
        """Get retarded Green function at specified energy.

        If 'inverse' is True, the inverse Green function is returned (faster).
        """
        if energy != self.energy:
            self.energy = energy
            z = energy + self.eta * 1.j

            if self.S is None:
                self.Ginv[:] = 0.0
                self.Ginv.flat[:: len(self.S) + 1] = z
            else:
                self.Ginv[:] = z
                self.Ginv *= self.S
            self.Ginv -= self.H

            for selfenergy in self.selfenergies:
                self.Ginv -= selfenergy.retarded(energy)

        if inverse:
            return self.Ginv
        else:
            return np.linalg.inv(self.Ginv)

    def recursive(self, energy):

        if energy != self.energy:
            self.energy = energy
            z = energy + self.eta * 1.j

            mat_list_ii = []
            mat_list_ij = []
            mat_list_ji = []

            h_list_ij, s_list_ij = self.hs_list_ij
            for h_ij, s_ij in zip(h_list_ij,
                                  s_list_ij):

                mat_list_ij.append(z * s_ij - h_ij)
                mat_list_ji.append(z * s_ij.T.conj() - h_ij.T.conj())

            h_list_ii, s_list_ii = self.hs_list_ii
            for h_ii, s_ii in zip(h_list_ii,
                                  s_list_ii):
                mat_list_ii.append(z * s_ii - h_ii)

            # from IPython import embed
            # embed()
            mat_list_ii[0]  -= self.selfenergies[0].retarded(energy)
            mat_list_ii[-1] -= self.selfenergies[1].retarded(energy)

            self.g_1N = recursive_gf(mat_list_ii, mat_list_ij, mat_list_ji)

        return self.g_1N

    def calculate(self, energy, sigma):
        """XXX is this really needed"""
        ginv = energy * self.S - self.H - sigma
        return np.linalg.inv(ginv)

    def apply_retarded(self, energy, X):
        """Apply retarded Green function to X.

        Returns the matrix product G^r(e) . X
        """
        return np.linalg.solve(self.retarded(energy, inverse=True), X)

    def dos(self, energy):
        """Total density of states -1/pi Im(Tr(GS))"""
        if self.S is None:
            return -self.retarded(energy).imag.trace() / np.pi
        else:
            GS = self.apply_retarded(energy, self.S)
            return -GS.imag.trace() / np.pi

    def pdos(self, energy):
        """Projected density of states -1/pi Im(SGS/S)"""
        if self.S is None:
            return -self.retarded(energy).imag.diagonal() / np.pi
        else:
            S = self.S
            SGS = np.dot(S, self.apply_retarded(energy, S))
            return -(SGS.diagonal() / S.diagonal()).imag / np.pi

    def take_bfs(self, bfs, apply):
        h_pp, s_pp, c_mm = CoupledHamiltonian.take_bfs(self, bfs, apply)
        if apply:
            self.Ginv = np.empty(self.H.shape, complex)
        return h_pp, s_pp, c_mm



# from .calculators import TransportCalculator
# __all__ = ['TransportCalculator']
import numpy as np
from scipy import linalg

xp = np
xla = linalg

try:
    import cupy as cp
except:
    pass
else:
    xp = cp
    xla = cp.linalg


#try:
#    import _cppmodule as _cpp
#except ImportError as e:
#    if "GLIBCXX" in str(e):
#        msg = ("Cannot import __cppmodule")
#        raise ImportError(msg).with_traceback(e.__traceback__)
#    else:
#        raise



import numpy as np
from functools import reduce

from ase.data import covalent_radii
from ase.neighborlist import NeighborList
import ase.neighborlist


def add_hydrogens(atoms):
    """Add hydrogens to organic structure."""
    cov_radii = [covalent_radii[a.number] for a in atoms]
    nl = NeighborList(cov_radii, bothways = True, self_interaction = False)
    nl.update(atoms)

    need_a_H = []
    for a in atoms:
        nlist=nl.get_neighbors(a.index)[0]
        if len(nlist)<3:
            if a.symbol=='C':
                need_a_H.append(a.index)

    print("Added missing Hydrogen atoms: ", need_a_H)

    dCH=1.1
    for a in need_a_H:
        vec = np.zeros(3)
        indices, offsets = nl.get_neighbors(atoms[a].index)
        for i, offset in zip(indices, offsets):
            vec += -atoms[a].position +(atoms.positions[i] + np.dot(offset, atoms.get_cell()))
        vec = -vec/np.linalg.norm(vec)*dCH
        vec += atoms[a].position
        htoadd = ase.Atom('H',vec)
        atoms.append(htoadd)


def apply_order(atoms, indices):
    """Permute indices of atoms inplace."""
    for name, a in atoms.arrays.items():
        atoms.arrays[name] = a[indices]


def sort(atoms, order='xyz', round=1):
    """Sort atoms by positions.

    Args:
        atoms : (ase.Atoms object)
        order : (str) any permutation of 'xyz'
            i.e. 'xyz' means first along 'x', then along 'y'
            and eventually along 'z'.
        round : (int) 
            round atomic position to `round` digits.
    
    Returns:
        sorted atoms.
    """
    positions = atoms.positions.round(round)
    i = order.index('x')
    j = order.index('y')
    k = order.index('z')
    indices = np.lexsort((positions[:,k],positions[:,j],positions[:,i]))
    return atoms[indices]
    

def expand(atoms, n_uc, n_rep, direction='x', sides=[0, 1], round=1):
    """Expand atoms along one direction with terminating atoms.

    Args:
        atoms : (ase.Atoms object)
        n_uc : (int)
            Number of (unit-cell) atoms at the termini to
            include in the expansion.
        n_rep : (int)
            Number of times to repeat the subset of `n_uc` atoms.
        direction : (char)
            Direction along which to perform the expansion.
        sides : (int, list, tuple)
            Whether to expand to the left, right of both ends.
            left  only  - 0
            right only  - 1
            left and right - [0,1]
        round : (int) see sort
    """
    if not atoms.cell.orthorhombic:
        raise NotImplementedError('Cell must be orthorhombic')

    if isinstance(sides, int):
        sides = [sides]

    dirs = [0,1,2]
    d = dirs.pop('xyz'.index(direction))
    order = str(reduce(str.__add__, ['xyz'[i] for i in [d]+dirs])) 
    atoms = sort(atoms, order, round)
    
    cell = atoms.cell[d,d] 
    uc_cell = (atoms[0].position[d] + cell) - atoms[-n_uc].position[d]
    
    for n in range(n_rep):

        uc_ldl = atoms[-n_uc:]
        uc_ldr = atoms[:n_uc]

        uc_ldr.positions[:,d] += cell
        uc_ldl.positions[:,d] -= cell

        if 0 in sides:
            atoms = uc_ldl + atoms
            atoms.cell[d,d] += uc_cell
        if 1 in sides:
            atoms = atoms + uc_ldr
            atoms.cell[d,d] += uc_cell

        atoms.center()
        atoms = sort(atoms, direction)
    return atoms



import _thread

from transport import xp
from .solvers.recursive import get_mat_lists, \
                               recursive_gf

class Tcalc:
    """
    Class to compute T_e concurrently between two threads
    """
    def __init__(self, greenfuncition, energies, T_e):
        self.greenfuncition = greenfuncition
        self.selfenergies = self.greenfuncition.selfenergies
        self.energies = energies
        self.T_e = T_e
        self.sigma_L = None
        self.sigma_R = None
        self.bcs_lock = _thread.allocate_lock()
        self.rec_lock = _thread.allocate_lock()
        #
        self.rec_lock.acquire()

    def bcs(self):

        for energy in self.energies:

            self.bcs_lock.acquire()
            self.sigma_L = self.selfenergies[0].retarded(energy)
            self.sigma_R = self.selfenergies[1].retarded(energy)
            self.rec_lock.release()


    def rec(self):

        for e, energy in enumerate(self.energies):

            self.rec_lock.acquire()
            sigma_L = xp.asarray(self.sigma_L) #local
            sigma_R = xp.asarray(self.sigma_R) #local
            self.bcs_lock.release()

            gf = self.greenfuncition
            z = energy + gf.eta * 1.j
            mat_lists = get_mat_lists(z, gf.hs_list_ii, gf.hs_list_ij,
                               gf.hs_list_ji, sigma_L, sigma_R)
            lambda1_11 = 1.j * (sigma_L - xp.conj(sigma_L.T))
            lambda2_NN = 1.j * (sigma_R - xp.conj(sigma_R.T))
            gr_1N = recursive_gf(*mat_lists)
            ga_1N = xp.conj(gr_1N.T)
            self.T_e[e] = xp.trace(lambda1_11.dot(gr_1N).dot(lambda2_NN).dot(ga_1N)).real



import numpy as np


# with open('selfenergy','w') as fp:
#     m, n = Sigret.shape[1:]
#     for energy,selfenergy in zip(energies,Sigret):
#         fp.write('{:.9}'.format(energy))
#         fp.write('\n')
#         for i in range(3,4):
#             for j in range(3,4):
#                 fp.write('{:.9} {:.9} '.format(selfenergy[i,j].real,selfenergy[i,j].imag))
#             fp.write('\n')
#
# with open('selfenergy','w') as fp:
#     m, n = Sigret.shape[1:]
#     for energy,selfenergy in zip(energies,Sigret):
#         fp.write('{:.9}'.format(energy))
#         fp.write('\n')
#         fp.write('{:.9} {:.9} '.format(selfenergy[3,3].imag, selfenergy[3,3].real))
#         fp.write('\n')
#
# with open('selfenergy-mats-{}'.format(beta),'w') as fp:
#     m, n = Sigret.shape[1:]
#     for n,selfenergy in zip(ns,Sig_mats):
#         w_n = np.pi/beta * (2*n + 1)
#         fp.write('{:.9}'.format(w_n))
#         fp.write('\n')
#         fp.write('{:.9} {:.9} '.format(selfenergy[3,3].imag, selfenergy[3,3].real))
#         fp.write('\n')

def load_LSigma(filename):

    Sigmas = np.loadtxt(filename)
    Sigmas.T[[1,2]] = Sigmas.T[[2,1]]
    Sigmas = Sigmas.view('f8,c16')
    Sigmas = np.split(Sigmas,2)

    return Sigmas

def dump_variables(prefix, *variables, dir='.'):

    import os
    import itertools

    for var in variables:

        def variablename(var):
            return [tpl[0] for tpl in
            filter(lambda x: var is x[1], globals().items())]

        name = variablename(var)
        if len(name)>1:
            raise Warning('Multiple names for variable',name)
        name = name[0]
        file = os.path.join(dir,prefix)+'_'+name+'.pickle'

        pickle.dump(var, open(file,'wb'), 2)


def get_matsubara(w, Sigma, N_mats, beta, S):
#     Nw = len(w)
    Nw, n, m = Sigma.shape
    w_mats = np.pi/beta * (2*np.arange(N_mats) + 1)
    Re_mats = np.zeros((N_mats,n,m),float)
    Im_mats = np.zeros((N_mats,n,m),float)
    A_wmm = Sigma.imag/np.pi
    w_wmm = S[None, :] * w[:, None, None]
    w2_wmm = S[None, :] * (w**2)[:, None, None]
    w_mats_wmm = S[None, None, :] * w_mats[:, None, None, None]
    w2_mats_wmm = S[None, None, :] * (w_mats**2)[:, None, None, None]
    print(A_wmm.shape, w_wmm.shape, w_mats_wmm.shape)
    for i in range(N_mats):
        Re_mats[i] = np.trapz(A_wmm*w_wmm/(w2_wmm+w2_mats_wmm[i]),w,axis=0)
        Im_mats[i] = np.trapz(A_wmm*w_mats_wmm[i]/(w2_wmm+w2_mats_wmm[i]),w,axis=0)
    return Re_mats + 1j * Im_mats



from qtpyt.blocktridiag.btmatrix import (empty_buffer, BTMatrix)

from qtpyt.blocktridiag import solvers

import numpy as np


class GreenFunction:
    """Block Tridiagonal Green's Function.
    
    Allows to compute observables for a scattering region
    embedded in an environment described by selfenergies.
    The Hamiltonian and the Overlap are given as two lists
    containing the diagonal and upper diagonal block-matrices.

    Args:
        hs_list_ii : (list(xp.ndarray(complex, shape=(2,m,m))), len=N)
            List of block matrices on the diagonal. 

                hs_list_ii[i][0] = Hamilton-block[0,0]
                hs_list_ii[i][1] = Overlap-block[0,0]
            
        hs_list_ii : (list(xp.ndarray(complex, shape=(2,m,m))), len=N-1)
            List of block matrices on the upper diagonal. 

                hs_list_ij[i][0] = Hamilton-block[0,1]
                hs_list_ij[i][1] = Overlap-block[0,1]

        selfenergies : (list(block, indices, selfenergy))
            List of selfenergies that couple to this Hamiltonian.

                Hamiltonian[block, block][indices] << selfenergy

        solver : ('str')
            One of the solvers available in Solver given as string. 
            Used to compute selected parts of the Green's function.
            (i.e. invert the matrix.)

    """
    def __init__(self, hs_lists_ii, hs_lists_ij, selfenergies=[], eta=1e-5, solver='spectral'):

        self.hs_list_ii = hs_lists_ii
        self.hs_list_ij = hs_lists_ij

        self.selfenergies = selfenergies
        self.energy = None
        self.eta = eta

        # NOTE this block is called before
        # constructing H and S (see below)
        # to allow offloading the blocks to GPU.
        Solver = solvers.__dict__.get(solver.capitalize(), None)
        assert Solver is not None, "Invalid solver."
        self.solver = Solver(self)

        # Here the Hamiltonian and Overlap may be on GPU.
        def _fill(buffer, i):
            buffer.m_qii[0] = self.hs_list_ii[0][i]
            for q in range(1,buffer.N):
                buffer.m_qii[q] = self.hs_list_ii[q][i]
                buffer.m_qij[q-1] = self.hs_list_ij[q-1][i]
                buffer.m_qji[q-1] = self.hs_list_ij[q-1][i].T

        S = empty_buffer(len(hs_lists_ii))
        H = empty_buffer(len(hs_lists_ii))
        
        _fill(S, 1)
        _fill(H, 0)

        self.S = BTMatrix(S.m_qii, S.m_qij, S.m_qji)
        self.H = BTMatrix(H.m_qii, H.m_qij, H.m_qji)

    def get_transmission(self, energy):
        return self.solver.get_transmission(energy)

    def retarded(self, energy, inverse=False):
        return self.solver.retarded(energy)

    def get_dos(self, energy):
        G = self.retarded(energy)
        return - 1/np.pi * - G.dottrace(self.S).imag

    def get_pdos(self, energy):
        G = self.retarded(energy)
        return - 1/np.pi * - G.dotdiag(self.S).imag

    # def add_screening(self, V):
    #     if not hasattr(self, 'V'):
    #         self.V = np.zeros(self.nbf)
    #     assert V.size == self.nbf
    #     #Add screening and remove (if exists) current.
    #     h_qii = self.hs_list_ii[0]
    #     if sum(self.V) != 0:
    #         add_diagonal(h_qii, - self.V)
    #     self.V[:] = V
    #     add_diagonal(h_qii, self.V)


    # def remove_screening(self):
    #     h_qii = self.hs_list_ii[0]
    #     add_diagonal(h_qii, -self.V)
    #     self.V[:] = 0.


import numpy as np
from numba import njit
from scipy.spatial.distance import cdist
from functools import partial


from ase.dft.kpoints import get_monkhorst_pack_size_and_offset


def monkhorst_pack(size, offset=(0.,0.,0.)):
    """K-points from Monkhorst-Pack."""
    kpts = np.indices(size).reshape(len(size), -1).T
    kpts = (kpts + 0.5) / size - 0.5 + offset
    return kpts


def _get_dirs(direction='x'):
    """Get indices parallel and transverse directions.
    
    Args:
        direction : (str) parallel direction. 
            One of 'x', 'y' or 'z'
    """
    dirs_t = [0,1,2]
    dir_p = dirs_t.pop('xyz'.index(direction))
    return dir_p, dirs_t


@njit
def _index(array, item):
    """Find item in array."""
    for idx, val in enumerate(array):
        if val == item:
            return idx


def map_ik2k(kpts, k2ik):
    """Get mapping from irreducible (ikpts) to reducible k-points (kpts).
    
    Args:
        kpts : (np.ndarray) k-points
        k2ik : map from kpts to ikpts such that kpts[k2ik] = ikpts

    Returns:
        ik2k : map from ikpts to kpts.
        time_reversal : whether item of map must be time-reversed.    
    """
    # Find if there's a minimum distance with time-reversed k-point.
    # distances = cdist(kpts, np.dot(kpts, -np.eye(3).T)) % 1
    distances = cdist(kpts%1, -kpts%1) #% 1
    equivalent = np.where(distances.min(1)<1e-7,distances.argmin(1),-1)
    ik2k = np.empty(kpts.shape[0],int)
    ik2k[k2ik] = np.arange(len(k2ik))
    diff = np.setdiff1d(range(kpts.shape[0]), k2ik)
    for idx, val in enumerate(diff):
        try:
            ik2k[val] = _index(k2ik, equivalent[val])
        except:
            raise IndexError
    # ik2k[diff] = [_index(k2ik, item) for item in equivalent[diff]]
    time_reverse = np.zeros(kpts.shape[0],bool)
    time_reverse[diff] = True
    return ik2k, time_reverse


def apply_inversion_symm(kpts, direction='x'):
    """Reduce inversion (time reversal) symmetry along direction.
    
    Args:
        kpts : (np.ndarray) k-points
        direction : time-reverse direction

    Returns:
        ikpts : (np.ndarray) irreducible k-points
        k2ik : mapping from kpts to ikpts, i.e. kpts[k2ik] = ikpts
        ik2k : map from ikpts to kpts.
        time_reversal : whether item of map must be time-reversed.
    
    Example:
        kpts = monkhorst-pack(size)
        ikpts, k2ik, ik2k, time_reverse = apply_inversion_symm(kpts)
        A :: k-matrix at ikpts
        B :: k-matrix at kpts
        B = A[ik2k]
        B[time_reverse] = np.transpose(B[time_reverse],axes=(0,2,1))
    """
    p_dir = 'xyz'.index(direction)
    ikpts = []
    k2ik = []
    for bzk_index, bzk_k in enumerate(kpts):
        try:
            if bzk_k[np.nonzero(bzk_k)[0][p_dir]] > 0:
                ikpts.append(bzk_k)
                k2ik.append(bzk_index)
            else:
                continue
        # zero case
        except IndexError:
            ikpts.append(bzk_k)
            k2ik.append(bzk_index)
    ikpts = np.array(ikpts)
    k2ik = np.array(k2ik)
    ik2k, time_reverse = map_ik2k(kpts, k2ik)
    return ikpts, k2ik, ik2k, time_reverse


def build_partial_kpts(size, offset=(0.,0.,0.), direction='x'):
    """Build parallel and transverse k-points. 
    
    Args:
        size, offset : Monkhorst-Pack.
        direction : parallel direction.

    Returns:
        kpts_p : parallel k-points
        kpts_t : transverse k-points 
    """
    size = np.asarray(size)
    offset = np.asarray(offset)

    dir_p, dirs_t = _get_dirs(direction)

    size_p = np.ones(3,int)
    size_p[dir_p] = size[dir_p]
    offset_p = np.zeros(3)
    offset_p[dir_p] = offset[dir_p]
    
    kpts_p = monkhorst_pack(size_p,offset_p)

    size_t = np.ones(3,int)
    size_t[dirs_t] = size[dirs_t]
    offset_t = np.zeros(3)
    offset_t[dirs_t] = offset[dirs_t]

    kpts_t = monkhorst_pack(size_t,offset_t)

    return kpts_p, kpts_t


def fourier_sum(A, kpts, R, out=None):
    '''K-space Fourier sum.
    
    Args:
        A : (np.ndarray) matrices in k-space.
        kpts : (1D or 2D) k-points at which A's have been evaluated.
        R : (1D or 2D) lattice vectors at which A's will be transformed.
    '''
    # np.asarray avoids copy
    kpts = np.asarray(kpts)
    R = np.asarray(R)
    if kpts.ndim<2:
        kpts = kpts[None,:]
    squeeze = False
    if R.ndim<2:
        squeeze = True
        R = R[None, :]
    shape = (R.shape[0],) + A.shape[1:]
    if out is None:
        out = np.empty(shape, dtype=A.dtype)
    if out.ndim<3:
        out = out[None,:]

    out.shape = (out.shape[0], -1)
    A.shape = (A.shape[0], -1)
    phase  = np.exp(2.j * np.pi * np.dot(R, kpts.T)) / kpts.shape[0]
    np.sum(phase[...,None] * A[None,...], axis=1, out=out)
    A.shape = (kpts.shape[0],) + shape[1:]
    out.shape = shape

    if squeeze:
        return out[0]
    return out


def partial_bloch_sum(A, kpts, R, kpts_p, kpts_t, out=None):
    """Partial Bloch sum along parallel direction.

    Args:
        A : (np.ndarray) matrices in k-space.
        kpts : (2D) k-points at which A's have been evaluated.
        R : (1D or 2D) lattice vectors at which A's will be transformed.
        kpts_p : parallel k-points
        kpts_t : transverse k-points
    """
    distance = partial(np.linalg.norm, axis=1)
    
    R = np.asarray(R)
    squeeze = False
    if R.ndim<2:
        squeeze = True
    R = np.array(R, ndmin=2)
    if out is None:
        out = np.empty((kpts_t.shape[0],R.shape[0]) + A.shape[1:], A.dtype)
    
    for j, kpt_t in enumerate(kpts_t):
        _kpts = kpts_p + kpt_t
        A_t = []
        for k in _kpts:
            try:
                i = np.where(distance(k-kpts)<1e-7)[0][0]
                A_t.append(A[i])
            except IndexError:
                # kpts are irreducible representation (ikpts).
                # look for time-reversed k-point k = -k + N.
                i = np.where(distance(((-k)%1)-kpts%1)<1e-7)[0][0]
                A_t.append(A[i].T)
        fourier_sum(np.array(A_t), _kpts, R, out[j])
    
    if squeeze:
        return out[:,0,:,:]
    return out


def build_lattice_vectors(size):
    """Lattice vectors in scaled coordinates."""
    return np.indices(size).reshape(len(size), -1).T


def unfold_kpts(Kpts, kpts):
    '''Unfold coarse mesh Kpts into fine mesh kpoints 
    sampled with Monkhorst-Pack mesh kpts_size.

    Args:
        Kpts : (np.ndarray, ndimn=2)
            The coarse sampled k-points.
        kpts : (np.ndarray, ndimn=2)
            The fine sampled k-points
        kpts_size : the size of the fine kpts mesh.

    Returns:
        k_groups : The groups of kpoints indices that must be
            unfolded to fold into Kpts.
    '''
    distance = partial(np.linalg.norm, axis=1)
    Ksize, Koffset = get_monkhorst_pack_size_and_offset(Kpts)
    ksize, koffset = get_monkhorst_pack_size_and_offset(kpts)

    #assert np.allclose(Koffset, koffset), f'Incompatible mesh grids.\
#They have different offsets {Koffset} vs. {koffset}. \
#    Force including the gamma point to match the offsets.'

    Nr = ksize//Ksize
    assert np.allclose(Nr, Nr.astype(int)), f'\
Incompatible mesh grids. Fine mesh with size {ksize} is \
not a multiple of coarse mesh with size {Ksize}.'

    k_groups = []
    for K in Kpts:
        group_indices = []
        for i,j,k in np.ndindex(tuple(Nr)):
            kpt = (K%1+(i,j,k))/Nr
            idx = np.where(distance(kpt-kpts%1)<1e-7)[0][0]
            group_indices.append(idx)
        k_groups.append(group_indices)
    
    return k_groups

    





from qtpyt import xp
from qtpyt.blocktridiag.recursive import (dyson_method, spectral_method, coupling_method_N1)
from qtpyt.blocktridiag.btmatrix import (empty_buffer, BTMatrix)

from qtpyt.abc._kernels import (_mul,_mulc,get_lambda,dagger)

class Solver:
    """The parent class of a generic solver.

    The principal task of Solver is to setup the inverse Green's
    function matrix which is then inverted by the child solver.
    It also offloas the Hamiltonian and Overlap matrices to GPU.

    """
    def __init__(self, greenfunction, method):
        N = len(greenfunction.hs_list_ii)
        self.greenfunction = greenfunction
        self.method = method
        
        # Possibly offload to GPU.
        gf = self.greenfunction
        gf.hs_list_ii[0] = xp.asarray(gf.hs_list_ii[0])
        for q in range(1,N):
            gf.hs_list_ii[q] = xp.asarray(gf.hs_list_ii[q])
            gf.hs_list_ij[q-1] = xp.asarray(gf.hs_list_ij[q-1])

        self.Ginv = empty_buffer(N)
        self.gammas = [None for _ in range(len(greenfunction.selfenergies))]

    def inv(self, Ginv, *args, **kwargs):
        return self.method(Ginv.m_qii, Ginv.m_qij, Ginv.m_qji, *args, **kwargs)

    def get_Ginv(self, energy):
        gf = self.greenfunction
        z = energy + gf.eta * 1.j
        for q, (h_ii, s_ii) in enumerate(gf.hs_list_ii):
            self.Ginv.m_qii[q] = _mul(z,h_ii,s_ii)
        # Upper and lower diagonals.
        for q, (h_ij, s_ij) in enumerate(gf.hs_list_ij):
            self.Ginv.m_qij[q] = _mul(z,h_ij,s_ij)
            self.Ginv.m_qji[q] = _mulc(z,h_ij.T,s_ij.T)
        # Add selfenergies
        for i, (block, selfenergy) in enumerate(gf.selfenergies):
            # Possibly offload to GPU.
            sigma = xp.asarray(selfenergy.retarded(energy))
            self.Ginv.m_qii[block] -= sigma
            self.gammas[i] = get_lambda(sigma)

        return self.Ginv

    def get_transmission(self, energy):
        raise NotImplementedError('{self.__class__.__name__} does not implement transmission.')

    def get_retarded(self, energy):
        raise NotImplementedError('{self.__class__.__name__} does not implement greenfunction.')

    def get_spectrals(self, energy):
        raise NotImplementedError('{self.__class__.__name__} does not implement spectrals.')


class Spectral(Solver):
    def __init__(self, greenfunction):
        super().__init__(greenfunction, spectral_method)
        self.A1 = None
        self.A2 = None

    def get_spectrals(self, energy):
        gf = self.greenfunction
        if energy != gf.energy:
            gf.energy = energy
            Ginv = self.get_Ginv(energy)
            self.A1 = None
            self.A2 = None
            A1, A2 = self.inv(Ginv, self.gammas[0], self.gammas[1])
            self.A1 = BTMatrix(*A1)
            self.A2 = BTMatrix(*A2)
        return self.A1, self.A2

    def get_transmission(self, energy):
        A2 = self.get_spectrals(energy)[1]
        gamma_L = self.gammas[0]
        T_e =  gamma_L.dot(A2[0,0]).real.trace()
        return T_e

    def get_retarded(self, energy):
        A1, A2 = self.get_spectrals(energy)
        return A1+A2


class Coupling(Solver):
    def __init__(self, greenfunction):
        super().__init__(greenfunction, coupling_method_N1)
    
    def get_transmission(self, energy):
        gf = self.greenfunction
        gf.energy = energy
        Ginv = self.get_Ginv(energy)
        g_N1 = self.inv(Ginv)
        gamma_L = self.gammas[0]
        gamma_R = self.gammas[1]
        T_e = xp.einsum('ij,jk,kl,lm->im',gamma_R,g_N1,
                        gamma_L,dagger(g_N1),optimize=True).real.trace()
        return T_e


class Dyson(Solver):
    def __init__(self, greenfuncition, trans=True):
        super().__init__(greenfuncition, dyson_method)
        self.G = None
        self.g_1N = None
        self.trans = trans

    def get_retarded(self, energy):
        gf = self.greenfunction
        if energy != gf.energy:
            gf.energy = energy
            Ginv = self.get_Ginv(energy)
            self.G = None
            if self.trans:
                g_1N, G = self.inv(Ginv, trans=True)
                self.g_1N = g_1N
                self.G = BTMatrix(*G)
            else:
                G = self.inv(Ginv, trans=False)
                self.G = BTMatrix(*G)
        return self.G
    
    def get_transmission(self, energy):
        _ = self.get_retarded(energy)
        gamma_L = self.gammas[0]
        gamma_R = self.gammas[1]
        T_e = xp.einsum('ij,jk,kl,lm->im',gamma_L,self.g_1N,
            gamma_R,dagger(self.g_1N),optimize=True).real.trace()
        return T_e